local StateContainerFactory = {}


function StateContainerFactory.createContainer()

    local StateContainer = {}
    StateContainer.ERROR_INVALID_FILE_CONTENTS = -1
    StateContainer.ERROR_HASH_MISS_MATCH       = -2
    StateContainer.ERROR_STATE_FILE_NOT_FOUND  = -3

    local crypto = require("framework.crypto")
    local json   = require("framework.json")

    local encodeSign    = "=QP="
    local stateFilename = "state.txt"
    local eventListener = nil
    local secretKey     = nil
    local disableEncryption = false
    local _currentData  = nil

    local encryptSign   = "goat_"

    local function isEncodedContents_(contents)
        return string.sub(contents, 1, string.len(encodeSign)) == encodeSign
    end

    local function encrypt_(data)
        if not disableEncryption then
            local cipher = 'aes-128-cfb'
            local text = crypto.pad(data)
            local key = runaway.GameConfig:getAESJSONKey()
            local iv = runaway.GameConfig:getAESJSONIV()
            local enc = crypto.encodeBase64(crypto.encrypt(cipher, text, key, iv))
            return encryptSign..enc
        else
            return data
        end
    end

    local function decrypt_(data)
        if data:sub(1, #encryptSign) ~= encryptSign then
            return data
        else
            data = data:sub(#encryptSign + 1)
        end
        local cipher = 'aes-128-cfb'
        local text = crypto.decodeBase64(data)
        local key = runaway.GameConfig:getAESJSONKey()
        local iv = runaway.GameConfig:getAESJSONIV()
        local dec = crypto.unpad(crypto.decrypt(cipher, text, key, iv))
        return dec
    end

    local function encode_(values)
        local s = json.encode(values)
        local hash = crypto.md5(s..secretKey)
        local contents = json.encode({h = hash, s = s})
        return encodeSign..contents
    end

    local function decode_(fileContents)
        local contents = string.sub(fileContents, string.len(encodeSign) + 1)
        local j = json.decode(contents)

        if type(j) ~= "table" then
            echoError("StateContainer.decode_() - invalid contents")
            return {errorCode = StateContainer.ERROR_INVALID_FILE_CONTENTS}
        end

        local hash,s = j.h, j.s
        local testHash = crypto.md5(s..secretKey)
        if testHash ~= hash then
            echoError("StateContainer.decode_() - hash miss match")
            return {errorCode = StateContainer.ERROR_HASH_MISS_MATCH}
        end

        local values = json.decode(s)
        if type(values) ~= "table" then
            echoError("StateContainer.decode_() - invalid state data")
            return {errorCode = StateContainer.ERROR_INVALID_FILE_CONTENTS}
        end

        return {values = values}
    end

    ----------------------------------------

    function StateContainer.init(eventListener_, containerName_, secretKey_, disableEncryption_)
        -- if type(eventListener_) ~= "function" then
        --     echoError("StateContainer.init() - invalid eventListener")
        --     return false
        -- end

        local function defaultListener(event)
            return event.values
        end
        StateContainer.eventListener = eventListener_  or defaultListener

        if type(containerName_) == "string" then
            stateFilename = containerName_
        end

        if type(secretKey_) == "string" then
            secretKey = secretKey_
        end

        disableEncryption = disableEncryption_

        StateContainer.eventListener({
            name     = "init",
            filename = StateContainer.getStateContainerPath(),
            encode   = type(secretKey) == "string"
        })

        return true
    end

    function StateContainer.load()
        echoInfo("StateContainer.load() for "..stateFilename)
        local filename = StateContainer.getStateContainerPath()

        if not io.exists(filename) then
            echoInfo("StateContainer.load() - file \"%s\" not found", filename)
            return StateContainer.eventListener({name = "load", errorCode = StateContainer.ERROR_STATE_FILE_NOT_FOUND})
        end

        local contents = decrypt_(io.readfile(filename))
        echoInfo("StateContainer.load() - get values from \"%s\"", filename)

        local values
        local encode = false

        if secretKey and isEncodedContents_(contents) then
            local d = decode_(contents)
            if d.errorCode then
                return StateContainer.eventListener({name = "load", errorCode = d.errorCode})
            end

            values = d.values
            encode = true
        else
            values = json.decode(contents)
            if type(values) ~= "table" then
                echoError("StateContainer.load() - invalid data")
                return StateContainer.eventListener({name = "load", errorCode = StateContainer.ERROR_INVALID_FILE_CONTENTS})
            end
        end

        -- dump(values ,"StateContainer.load("..stateFilename..")")

        _currentData = values;
        return StateContainer.eventListener({
            name   = "load",
            values = values,
            encode = encode,
            time   = os.time()
        })
    end

    function StateContainer.save(newValues)
        echoInfo("StateContainer.save() for "..stateFilename)

        local values = StateContainer.eventListener({
            name   = "save",
            values = newValues,
            encode = type(secretKey) == "string"
        })
        if type(values) ~= "table" then
            echoError("StateContainer.save() - listener return invalid data")
            return false
        end

        local filename = StateContainer.getStateContainerPath()
        local ret = false
        if secretKey then
            local enc = encrypt_(encode_(values))
            ret = io.writefile(filename, enc)
        else
            local s = json.encode(values)
            local enc = encrypt_(s)
            if type(s) == "string" then
                ret = io.writefile(filename, enc)
            end
        end

        _currentData = newValues;
        return ret
    end

    function StateContainer.getCurrent()
        return _currentData
    end

    function StateContainer.delete()
        local filename = StateContainer.getStateContainerPath()
        if io.exists(filename) then
            echoInfo("Deleting StateContainer file : ".. filename)
            os.remove(filename)
        else
            echoInfo("Unable to find StateContainer file to delete : ".. filename)
        end
        _currentData = nil;
    end

    function StateContainer.getStateContainerPath()
        return string.gsub(device.writablePath, "[\\\\/]+$", "") .. "/" .. stateFilename
    end

    return StateContainer
end

return StateContainerFactory