-- THE MIGHTY TRANSLATATRON
-- Created by Aleksis Kavalieris 17/4/2014
local Localize = {}

local vault = {} -- every translation indexed by key

local fallbackLanguageCode = 'en'
local targetLanguageCode = nil

local fileOps = {}

function fileOps.parseFile(path)
    return cc.FileUtils:getInstance():getStringFromFile(path)
end

function fileOps.fileExists(path)
   return cc.FileUtils:getInstance():isFileExist(path)
end

function fileOps.parseJSONFile(path)
    local fileContents = fileOps.parseFile(path)
    local luaValue = json.decode(fileContents)
    return luaValue
end

function fileOps.splitPath(path)
    return string.match(path, "(.-)([^\\]-([^%.]+))$")
end

local function mergeTables(a, b)
    if not b then return a end
    for k,v in pairs(b) do
        a[k] = v
    end
    return a
end

local function makeLocalizationPath(stringsDirectory, languageCode)
    local separator = '/'
    if stringsDirectory:sub(-2, -1) == '/' then
        separator = ''
    end
    return stringsDirectory..separator..'Localized_'..languageCode..'.json'
end

local function getLocalizationFromFile(stringsDirectory, languageCode)
    local pathToLocalizationFile = makeLocalizationPath(stringsDirectory, languageCode)
    if fileOps.fileExists(pathToLocalizationFile) then
        return fileOps.parseJSONFile(pathToLocalizationFile)
    else
        print('Cannot find localization file at '..pathToLocalizationFile)
    end
end

function Localize.dumpVault()
    dump(vault)
end

function Localize.addKeyValueToVault(key,value)
    if not vault[key] then
        vault[key] = value
    end
end

function Localize.overrideKeyValueInVault(key,value)
    vault[key] = value
end

function Localize.addStrings(stringsDirectory, languageCode)
    targetLanguageCode = languageCode or 'en'

    local fallbackLocalization = getLocalizationFromFile(stringsDirectory, fallbackLanguageCode)
    local targetLocalization = getLocalizationFromFile(stringsDirectory, targetLanguageCode)

    local newStrings = mergeTables(fallbackLocalization, targetLocalization)
    vault = mergeTables(vault, newStrings)
end

function Localize.clearStrings()
    vault = {}
    print("Localisation strings cleared")
end

function Localize.keyExists(key)
    local result = vault[key]
    if result == nil then return false end
    return true
end

function Localize.localize(key)
    local result = vault[key]
    if result == nil then
        if not DEBUG_SUPPRESS_TRANSLATION_ERRORS then
            return 'NO TRANSLATION FOR KEY: '..key
        end
        return ""
    end
    return result
end

function Localize.format(key, ...)
    local args, order = {...}, {}
    local string = Localize.localize(key)
    local fmt = string:gsub('%%(%d+)%$', function(i)
        table.insert(order, args[tonumber(i)])
        return '%'
    end)
    return string.format(fmt, unpack(order))
end

function Localize.getRichText(key, formatOptions, ...)
    local function mergeStyles(styles, formatOptions)
        local result = {}
        for i = 1,#styles do
          local style = styles[i]
            if formatOptions[style] then
                for key,value in pairs(formatOptions[style]) do
                    result[key] = value
                end
            end
        end
        return result
    end

    local function getDesignations(element, formatOptions, currentStyles)
        local result = {}
        if element.name then
            if element.name == 'span' then
                if element.attr then
                    if element.attr.style then
                        table.insert(currentStyles, element.attr.style)
                    end
                end
            elseif element.name == 'img' then
                if element.attr then
                    if element.attr.style then
                        table.insert(currentStyles, element.attr.style)
                        current.style = mergeStyles(currentStyles, formatOptions)
                        table.remove(currentStyles)
                    end
                    if element.attr.src then
                        current = {}
                        print(element.attr.src)
                        current.src = formatOptions[element.attr.src].src
                        table.insert(result, current)
                        return result
                    end
                end
            end
        end
        for k = 1,#element do
          local v = element[k]
            if type(v) == 'string' then
                local current = {}
                current.text = v
                current.style = mergeStyles(currentStyles, formatOptions)
                table.insert(result, current)
            elseif type(v) == 'table' then
                designationPile = getDesignations(v, formatOptions, currentStyles)
                for k = 1,#designationPile do
                    local v = designationPile[k]
                    table.insert(result, v)
                end
            end
        end
        table.remove(currentStyles)
        return result
    end

    local formattedText = Localize.format(key, ...)
    local xmlTree = xml.parseLine(formattedText)
    local designations = getDesignations(xmlTree, formatOptions, {})

    local richTextResult = ccui.RichText:create()
    local colour, alpha, font, size, re

    for k = 1,#designations do
        local v = designations[k]
        local style = v.style or {}
        colour = style.colour or {255,255,255}
        alpha = style.alpha or 255
        font = style.font or 'Arial'
        size = style.size or 18
        text = v.text or ''
        imgsrc = v.src or ''

        if imgsrc ~= '' then
            re = ccui.RichElementImage:create(1, cc.c3b(colour[1],colour[2],colour[3]), alpha, imgsrc)
        else
            re = ccui.RichElementText:create(1, cc.c3b(colour[1],colour[2],colour[3]), alpha, text, font, size)
        end

        richTextResult:pushBackElement(re)
    end
    return richTextResult
end

return Localize
