--tmxParser.lua
--The built-in TMX XML importer in Cocos2d-x understandably does not support the development features
--of Tiled, so I will have to make a parser that does.

local RunawayTMXParser = {}

--Will print information about what the parser is doing
local debugPrint = false

--Width of tiles on the map
local tileWidth

--Height of tiles on the map
local tileHeight

--Number of tiles across the map
local mapWidth

--Number of tile up and down the map
local mapHeight

-- Functions for getting and setting single bits in lua numbers
local bitUtils = {}

function bitUtils.bit(p)
return 2 ^ (p - 1)  -- 1-based indexing
end

-- Typical call: if hasbit(x, bit(3)) then ...
function bitUtils.hasbit(x, p)
	return x % (p + p) >= p
end

-- Typical call: setbit(x, bit(3))
function bitUtils.setbit(x, p)
	return bitUtils.hasbit(x, p) and x or x + p
end

-- Typical call: clearbit(x, bit(3))
function bitUtils.clearbit(x, p)
	return bitUtils.hasbit(x, p) and x - p or x
end

local function splitString(str, delim, maxNb)
	if string.find(str, delim) == nil then
		return { str }
	end
	if maxNb == nil or maxNb < 1 then
      maxNb = 0   -- No limit
  end
  local result = {}
  local pat = "(.-)" .. delim .. "()"
  local nb = 0
  local lastPos
  for part, pos in string.gfind(str, pat) do
  	nb = nb + 1
  	result[nb] = part
  	lastPos = pos
  	if nb == maxNb then break end
  end
   -- Handle the last field
   if nb ~= maxNb then
   	result[nb + 1] = string.sub(str, lastPos)
   end
   return result
end

local function Strip_Control_Codes( str )
	local s = ""
	for i in str:gmatch( "%C+" ) do
		s = s .. i
	end
	return s
end

local function Strip_Control_and_Extended_Codes( str )
local s = ""
for i = 1, str:len() do
	if str:byte(i) >= 32 and str:byte(i) <= 126 then
		s = s .. str:sub(i,i)
	end
end
return s
end

--A reversed ipairs iterator
local function ripairs(t)
	idx={}
	for k,v in pairs(t) do
		if type(k)=="number" then idx[#idx+1]=k end
	end
	table.sort(idx)
	local function ripairs_it(t,_)
		if #idx==0 then return nil end
		k=idx[#idx]
		idx[#idx]=nil
		return k,t[k]
	end
	return ripairs_it, t, nil
end

local function printChildTagNames(data)
	for i = 1,#data.kids do
        local v = data.kids[i]
		if v.type == "element" then
			print(v.name)
			printChildTagNames(v)
		end
	end
end

local function findChildTagsWithName(data, name)
	local result = {}
	for i = 1,#data.kids do
        local v = data.kids[i]
		if v.type == "element" and v.name == name then
			v.index = i
			table.insert(result, v)
		end
	end
	return result
end

local function findAttributeWithName(data, name)
	for i = 1,#data.attr do
        local v = data.attr[i]
		if v.name == name and v.type == "attribute" then
			return v.value
		end
	end
	if debugPrint then
		print('Attribute '..name..' not found')
		return
	end
end

local function buildGIDMap(tilesets)
	local result = {}
	for i,tileset in pairs(tilesets) do
		local firstGID = tonumber(findAttributeWithName(tileset, "firstgid"))
		local tiles = findChildTagsWithName(tileset, "tile")
		for i,tile in pairs(tiles) do
			if tile.type == "element" then
				local localGID = tonumber(findAttributeWithName(tile, "id"))
				local actualGID = localGID + firstGID
				local imageTag = findChildTagsWithName(tile, "image")
				local textureSource = findAttributeWithName(imageTag[1], "source")
				result[actualGID] = textureSource
			end
		end
	end
	return result
end

local function clearHighBits(objectGID)
	cleanGID = objectGID
	cleanGID = bitUtils.clearbit(cleanGID, bitUtils.bit(32))
	cleanGID = bitUtils.clearbit(cleanGID, bitUtils.bit(31))
	cleanGID = bitUtils.clearbit(cleanGID, bitUtils.bit(30))
	return cleanGID
end

local function getFlipsAndGIDFromDataThing(dataThing) -- Can do rotation via flips
	local bit32 = bitUtils.hasbit(dataThing, bitUtils.bit(32)) -- bit32 stores horizontal flip
	local bit31 = bitUtils.hasbit(dataThing, bitUtils.bit(31)) -- bit31 stores vertical flip
	local bit30 = bitUtils.hasbit(dataThing, bitUtils.bit(30)) -- bit30 stores anti-diagonal flip, whatever that means
	local cleanGID = dataThing
	local hflip = false
	local vflip = false
	local dflip = false
	if bit32 then
		hflip = true
	end
	if bit31 then
		vflip = true
	end
	if bit30 then
		dflip = true
	end
	cleanGID = clearHighBits(cleanGID)
	return cleanGID, {h=hflip, v=vflip, d=dflip}
end

local function buildAllLayers(objects, layers, GIDMap)
	for i = 1,#layers do
        local layer = layers[i]
		local layerName = findAttributeWithName(layer, 'name')
		local data = findChildTagsWithName(layer, 'data')[1]
		local tiles = findChildTagsWithName(data, 'tile')
		for i = 1,#tiles do
        	local tile = tiles[i]
			local dirtyGID = tonumber(findAttributeWithName(tile, 'gid'))
			if dirtyGID ~= 0 then
				local GID, flippies = getFlipsAndGIDFromDataThing(dirtyGID)
				local textureSource = GIDMap[GID]
				local positionX = ((i - 1) % mapWidth) * tileWidth
				local positionY = (math.floor(((i - 1) / mapWidth) + 1) - mapHeight) * tileHeight
				positionX = positionX + tileWidth/2
				positionY = positionY - tileHeight/2
				local layerID = layer.index
				local object = {
					source = textureSource, 
					x = positionX, 
					y = -positionY, 
					r = 0, 
					f = flippies, 
					layer = layerID, 
					layerName = layerName
				}
				table.insert(objects, object)
			end
		end
	end
end

local function getPointsFromPolygon(object)
	local points = {}
	for _,kid in pairs(object.kids) do
		if kid.name == 'polygon' then
			for _,attributes in pairs(kid.attr) do
				for _,attributeName in pairs(attributes) do
					if attributeName ~= 'points' and
						attributeName ~= 'attribute' then
						local coords = splitString(attributeName, '%s')
						for i = 1,#coords do
        					local position = coords[i]
							local point = splitString(position, ',')
							point = cc.p(point[1], point[2])
							table.insert(points, point)
						end
					end
				end
			end
		end
	end
	return points
end

local function getConnectedRoomInfo(object)
	-- probably a better way of doing this...
	local propertyExists = false
	local connectedRooms = {}
	for _,kids in pairs(object.kids) do
		for _,kid in pairs(kids) do
			if type(kid) == 'table' then
				for _,properties in pairs(kid) do
					for _,prop in pairs(properties) do
						if type(prop) == 'table' then
							for _,value in pairs(prop) do
								for k,v in pairs(value) do
									if k == 'value' then
										if v ~= 'connectedRooms' then
											connectedRooms = splitString(v, ',')
										else
											propertyExists = true
										end
									end
								end
							end
						end
					end
				end
			end
		end
	end
	if propertyExists then
		return connectedRooms
	else
		return
	end
end

local function buildAllObjectGroups(objects, objectGroups, GIDMap)
	for i = 1,#objectGroups do
        local objectGroup = objectGroups[i]
		local objectGroupName = findAttributeWithName(objectGroup, 'name')
		print('objectGroupName', objectGroupName)
		local listOfObjects = findChildTagsWithName(objectGroup, 'object')
		for i,object in ripairs(listOfObjects) do
			local dirtyGID, width, height, GID, flippy, textureSource
			local points = {}
			if objectGroupName == "walls" or
				objectGroupName == "zones" or
				objectGroupName == "stencils" or
				objectGroupName == "expansions" or
				objectGroupName == "noswimzones" or
				objectGroupName == "rooms" then
				points = getPointsFromPolygon(object)
				connectionInfo = getConnectedRoomInfo(object)
			else
				dirtyGID = tonumber(findAttributeWithName(object, 'gid'))
				GID, flippy = getFlipsAndGIDFromDataThing(dirtyGID)
				textureSource = GIDMap[GID]
			end
			local objName = findAttributeWithName(object, 'name')
			local positionX = tonumber(findAttributeWithName(object, 'x'))
			local positionY = tonumber(findAttributeWithName(object, 'y')) - (mapHeight * tileHeight)
			local rotationString = findAttributeWithName(object, 'rotation')
			local rotation = tonumber(rotationString or 0)
			local layerID = objectGroup.index
			local guid = findAttributeWithName(object, 'uid')
			if guid then guid = 'u'..guid end

			local result = {
				source = textureSource, 
				x = positionX, 
				y = -positionY, 
				r = rotation, 
				f = flippy,
				guid = guid,
				layer = layerID, 
				layerName = objectGroupName, 
				points = points, 
				objName = objName, 
				connectedRooms = connectionInfo, 
				isObject = true
			}

			table.insert(objects, result)
		end
	end
end

local function buildObjectTable(data)
	local objectTable = {}

	local mapTagTable = findChildTagsWithName(data, "map")[1]

	tileWidth = tonumber(findAttributeWithName(mapTagTable, 'tilewidth'))
	tileHeight = tonumber(findAttributeWithName(mapTagTable, 'tileheight'))
	mapWidth = tonumber(findAttributeWithName(mapTagTable, 'width'))
	mapHeight = tonumber(findAttributeWithName(mapTagTable, 'height'))

	-- First build a map between GID values and image sources
	local tilesets = findChildTagsWithName(mapTagTable, "tileset")
	local GIDMap = buildGIDMap(tilesets)

	-- Then build all of the layers to objects
	local layers = findChildTagsWithName(mapTagTable, "layer")
	buildAllLayers(objectTable, layers, GIDMap)

	-- Then after that build all of the object groups to objects
	local objectGroups = findChildTagsWithName(mapTagTable, "objectgroup")
	buildAllObjectGroups(objectTable, objectGroups, GIDMap)

	if debugPrint then printChildTagNames(data) end
	return objectTable
end

local function readAll(file)
	local path = file
	if cc.FileUtils:getInstance():isFileExist(path) then
		print("File "..path.." actually exists!")
	else
		print("File "..path.." does not exist!")
	end
	return cc.FileUtils:getInstance():getStringFromFile(path)
end

--Eats a tmx file and spits out usable map data
function RunawayTMXParser.eat(file)
	local raw = readAll(file)
	local data = xml.decode(raw)
	objectTable = buildObjectTable(data)
	--iprint(objectTable)
	return objectTable
end

return RunawayTMXParser
