--conditionWatcher.lua

local conditionWatcher = {}

--local eventDispatcher = cc.Director:getInstance():getEventDispatcher()

function conditionWatcher:init()
	self._observers = {}
	self._conditions = {}
end


--	sets a named condition to a value
function conditionWatcher:setCondition(conditionName, value)
	assert(type(conditionName) == "string", "conditionName is not a string (conditionWatcher:setCondition)")
	-- if DEBUG_CONDITION_WATCHER then
		echoInfo("Condition set (%s)=>(%s)", conditionName, value)
	-- end
	self._conditions[conditionName] = value;
	self:_triggerUpdates()
	self:updateDisplayIndicators()
end

--	gets a named condition value
function conditionWatcher:getCondition(conditionName)
	assert(type(conditionName) == "string", "conditionName is not a string (conditionWatcher:getCondition)")
	return self._conditions[conditionName];
end

--	remove a condition observer to stop handling events 
function conditionWatcher:removeConditionObserver(observerName)
	assert(type(observerName) == "string", "observerName is not a string (conditionWatcher:removeConditionObserver)")
	self._observers[observerName] = nil;
end

--	get the state of a condition observer 
function conditionWatcher:getConditionObserverState(observerName)
	assert(type(observerName) == "string", "observerName is not a string (conditionWatcher:getConditionObserver)")
	local observer = self._observers[observerName];
	if observer then
		return conditionWatcher:_getConditionStatus(observer.conditionStates)
	end
	return "undefined"
end

--	creates a conditionObserver for various conditions
--	VonditionStates is of the form:
--		{ "condition1" = value, "condition2" = value }
--	The conditionObserver has the following methods to register trigger functions:
--		observer.onAllConditionsAchieved(func)
--		observer.onAllConditionsAchievedBroken(func)
--		observer.onAnyConditionAchieved(func)
--		observer.onAllConditionsFailed(func)
--	Observers can be in 1 of 3 statuses:
--		All conditions methods	(ALL)
--		No conditions met 		(SOME)
--		Some conditions met 	(NONE)
--	The following transitions trigger the listed functions:
--		NONE 		=> SOME/ALL		observer.onAnyConditionAchieved
--		NONE/SOME 	=> ALL 			observer.onAllConditionsAchieved
--		ALL 		=> SOME/NONE	observer.onAllConditionsAchievedBroken
--		ALL/SOME 	=> NONE			observer.onAllConditionsFailed
function conditionWatcher:createConditionObserver(observerName)
	assert(type(observerName) == "string", "observerName is not a string (conditionWatcher:createConditionObserver)")
	local observer = {}
	observer.conditionStates = {};
	observer.observerName = observerName;
	observer._onAllConditionsAchievedFunc = nil;
	observer._onAllConditionsAchievedBrokenFunc = nil;
	observer._onAnyConditionAchievedFunc = nil;
	observer._onAllConditionsFailedFunc = nil;
	function observer:onAllConditionsAchieved(func) 		self._onAllConditionsAchievedFunc = func		end
	function observer:onAllConditionsAchievedBroken(func) 	self._onAllConditionsAchievedBrokenFunc = func 	end
	function observer:onAnyConditionAchieved(func) 		 	self._onAnyConditionAchievedFunc = func 		end
	function observer:onAllConditionsFailed(func) 			self._onAllConditionsFailedFunc = func 			end
	function observer:addCondition(conditionName, expectedValue)
		if DEBUG_CONDITION_WATCHER then
			echoInfo("Adding new condition(%q) to observer(%q) => ", conditionName, self.observerName)
			echoInfo("Current observer Status = %q", self.currentStatus or "undefined")
		end
		self.conditionStates[conditionName] = expectedValue;

		self.currentStatus = conditionWatcher:_getConditionStatus(self.conditionStates)
		if DEBUG_CONDITION_WATCHER then
			echoInfo("new observer Status = %q", self.currentStatus)
		end
	end

	self._observers[observerName] = observer;

	return observer;
end





------------------------------------------------------------------------------------
--			internal methods
------------------------------------------------------------------------------------

local CONDITION_STATUS_ALL = "all"
local CONDITION_STATUS_SOME = "some"
local CONDITION_STATUS_NONE = "none"

function conditionWatcher:_getConditionStatus(conditionStates)
	local status = ""
	for conditionName, expectedValue in pairs(conditionStates) do
		local conditionValue = self:getCondition(conditionName)
		if (conditionValue == expectedValue) then
			if (status == "") then
				status = CONDITION_STATUS_ALL;
			elseif (status == CONDITION_STATUS_NONE) then
				status = CONDITION_STATUS_SOME;
			end
		else
			if (status == "") then
				status = CONDITION_STATUS_NONE;
			elseif (status == CONDITION_STATUS_ALL) then
				status = CONDITION_STATUS_SOME;
			end
		end
	end
	return status;
end


function conditionWatcher:_triggerUpdates()
	for observerName, observer in pairs(self._observers) do
		if (observer ~= nil) then
			local oldStatus = observer.currentStatus;
			local newStatus = self:_getConditionStatus(observer.conditionStates);
			if (oldStatus ~= newStatus) then
				if DEBUG_CONDITION_WATCHER then
					echoInfo("Observer [%s] transitioned from %s => %s", observerName, oldStatus, newStatus)
				end
				observer.currentStatus = newStatus;

				--	NONE 		=> SOME/ALL		observer.onAnyConditionAchieved
				if (oldStatus == CONDITION_STATUS_NONE) 
					and (newStatus == CONDITION_STATUS_SOME or newStatus == CONDITION_STATUS_ALL) then
						self:_safeExecuteFunction(observerName, "onAnyConditionAchieved", observer._onAnyConditionAchievedFunc);
				end

				--	NONE/SOME 	=> ALL 			observer.onAllConditionsAchieved
				if (oldStatus == CONDITION_STATUS_NONE or oldStatus == CONDITION_STATUS_SOME)
					and (newStatus == CONDITION_STATUS_ALL) then
						self:_safeExecuteFunction(observerName, "onAllConditionsAchieved", observer._onAllConditionsAchievedFunc);
				end

				--	ALL 		=> SOME/NONE	observer.onAllConditionsAchievedBroken
				if (oldStatus == CONDITION_STATUS_ALL) 
					and (newStatus == CONDITION_STATUS_SOME or newStatus == CONDITION_STATUS_NONE) then
						self:_safeExecuteFunction(observerName, "onAllConditionsAchievedBroken", observer._onAllConditionsAchievedBrokenFunc);
				end
				
				--	ALL/SOME 	=> NONE			observer.onAllConditionsFailed
				if (oldStatus == CONDITION_STATUS_ALL or oldStatus == CONDITION_STATUS_SOME)
					and (newStatus == CONDITION_STATUS_NONE) then
						self:_safeExecuteFunction(observerName, "onAllConditionsFailed", observer._onAllConditionsFailedFunc);
				end
			end
		end
	end
end

function conditionWatcher:_safeExecuteFunction(observerName, functionName, func)
	if (type(func) == "function") then
		if DEBUG_CONDITION_WATCHER then
			echoInfo("Observer Triggering %s.%s()", observerName, functionName)
		end
		func()
	end
end

--	give debug feedback of enabled conditions
function conditionWatcher:enableDisplayIndicators(enabled)
	--	if no parameter is supplied, then toggle it.
	if type(enabled) == "nil" then
		enabled = not self._displayIndicatorsEnabled
	end

	self._displayIndicatorsEnabled = enabled

	if enabled then
		self:updateDisplayIndicators()
	elseif self._indicatorLabel ~= nil then
		local label = self:getIndicatorLabel()
		if label ~= nil then
			SYSTEMLAYER:removeChild(label)
		end
	end
end

function conditionWatcher:getIndicatorLabel()
	if SYSTEMLAYER ~= nil then
		local taggedLabel = SYSTEMLAYER:getChildByTag(665)
		-- if taggedLabel == self._indicatorLabel then
			return taggedLabel
		-- end
	end
	return nil
end

function conditionWatcher:getConditionText()
	local labelText = "Conditions:"
	for k, v in pairs(self._conditions) do
		local value = v
		if type(v) == "boolean" then
			value = v and "YES" or "NO"
		end
		labelText = labelText .. "\n  " .. k .. "=>" .. value 
	end
	labelText = labelText .. "\nTriggers:"
	for observerName, observer in pairs(self._observers) do
		if (observer ~= nil) then
			labelText = labelText .. "\n  " .. observerName .. "=>" .. observer.currentStatus
		end
	end

	labelText = labelText .. "\npersistence is " .. (persistence.paused and "PAUSED" or "running")
	return labelText
end

function conditionWatcher:updateDisplayIndicators()
	if self._displayIndicatorsEnabled == true and SYSTEMLAYER ~= nil then
		local label = self:getIndicatorLabel()

		if label == nil then
			echoInfo("No indicator Label found, creating")
			label = cc.Label:createWithTTF("Indicators", FONT_ARIAL or FONT_GILLSANS_BOLD, 6, cc.size(display.width,0), cc.TEXT_ALIGNMENT_LEFT, cc.VERTICAL_TEXT_ALIGNMENT_TOP)

			label:setAnchorPoint(cc.p(0.5,0.5))
			label:setPosition(cc.p(display.cx, display.height * 0.8))
			label:enableShadow(SHADOW_COLOR, utils.convertShadowOffsetFromPointToSize(SHADOW_OFFSET_AS_PT), SHADOW_BLUR)
			label:setTag(665)
			SYSTEMLAYER:addChild(label)
			self._indicatorLabel = label
		end


		label:setString(self:getConditionText())
	end
end

function conditionWatcher:debugDump()
	echoInfo(self:getConditionText())
end


return conditionWatcher
