local NODE_UNCHANGED = "_DIFF_UNCHANGED"
local NODE_CHANGED   = "_DIFF_MODIFY"
local NODE_ADDED     = "_DIFF_INSERT"
local NODE_REMOVED   = "_DIFF_DELETE"
local DELETE_PREFIX  = "_DIFF_DELETE_"
local INCLUDE_GUID_LOOKUPS = false;     --  use Guid attribute lookups for Flutter
local GUID_FIELDS    = { "GUID", "flowerID"};


local diffPatch = {}

function isEmptyTable(t)
    if (type(t) == "table") then
        for _, __ in pairs(t) do
            return true;
        end
    end
    return true;
end

--  returns true if the table is keyed (instead of indexed)
function isKeyedTable(t)
    if (type(t) == "table") then
        for k,v in pairs(t) do
            if (type(k) ~= "number") then
                return true;
            end
        end
        -- if we have no items in the table then we can't correctly determine if it is keyed.
        return false;
    end
    return false;
end

--  returns true if the table is values only (no child tables)
function isTableValuesOnly(t)
    if (t ~= nil) then
        for k,v in pairs(t) do
            if (type(v) == "table") then
                return false;
            end
        end
    end
    return true;
end


--  handle special case where table values have "GUID" field
function isTableGuidKeyed(t)
    if (type(t) == "table") then
        for _, value in pairs(t) do
            if (type(value) == "table") then
                for __, field in pairs(GUID_FIELDS) do
                    if value[field] ~= nil then
                        return true;
                    end
                end
            end 
            return false;
        end
    end
    return false;
end
function getValueByGuid(t, guid)
    if (type(t) == "table") then
        for _, value in pairs(t) do
            for __, field in pairs(GUID_FIELDS) do
                if value[field] ~= nil and value[field] == guid then
                    return value;
                end
            end
        end
    end
    return nil;
end
function getValueGuid(value)
    for __, field in pairs(GUID_FIELDS) do
        if value[field] ~= nil then
            return value[field];
        end
    end
end

-- function table.contains(t, element)
--     if (t ~= nil) then
--         for _, value in pairs(t) do
--             if value == element then
--                 return true
--             end
--         end
--     end
--     return false
-- end




function walkTree(lhs, rhs, key, parentNode)
    local thisNode = {};
    if (parentNode.children == nil) then
        parentNode.children = {};
    end
    table.insert(parentNode.children, thisNode)
    thisNode.parent = parentNode;  
    thisNode.key = key;
    thisNode.lhs = lhs;
    thisNode.rhs = rhs;
    thisNode.name = (parentNode.name).."."..(key or ""); 
    thisNode.isGuidKeyed = false;

    if (lhs == nil) then
        thisNode.action = NODE_REMOVED;
        return;
    elseif (rhs == nil) then
        thisNode.action = NODE_ADDED;
    else
        thisNode.action = NODE_UNCHANGED;   --until otherwise determined
    end

    if (type(lhs) == "table" or type(rhs) == "table") then
        
        if (isKeyedTable(lhs) or isKeyedTable(rhs)) then

            --  iterate through the LHS, noting those items added.
            for key, value in pairs(lhs) do
                if (rhs == nil or rhs[key] == nil) then
                    walkTree(value, nil, key, thisNode);
                else
                    walkTree(value, rhs[key], key, thisNode);
                end
            end

            -- now iterate through RHS pulling out all removed items
            if (rhs ~= nil) then
                for key, value in pairs(rhs) do
                    if (lhs[key] == nil) then
                        --  this node is removed, so we don't bother 
                        walkTree(nil, rhs, key, thisNode)
                    end 
                end
            end

        elseif (isTableValuesOnly(lhs) or isTableValuesOnly(rhs)) then

            --  iterate through the LHS, noting those items added.
            for _, value in pairs(lhs) do
                if (not table.contains(rhs, value)) then
                    walkTree(value, nil, nil, thisNode)
                end
            end

            -- now iterate through RHS pulling out all removed items
            if (rhs ~= nil) then
                for _, value in pairs(rhs) do
                    if (not table.contains(lhs, value)) then
                        walkTree(nil, value, nil, thisNode)
                    end
                end
            end

        elseif INCLUDE_GUID_LOOKUPS and (isTableGuidKeyed(lhs) or isTableGuidKeyed(rhs)) then

            --  special case where table items include GUID value;
            thisNode.isGuidKeyed = true;
            print (" ");
            print ("***********************************************");
            print ("******************  WARNING  ******************");
            print ("***********************************************");
            print ("**found Guid based table: ".. thisNode.name);
            print ("**  Converting to Key'ed table!              **");
            print ("***********************************************");
            print (" ");

            for _, value in pairs(lhs) do
                local guid = getValueGuid(value);
                walkTree(value, getValueByGuid(rhs, guid), guid, thisNode)
            end

            -- now iterate through RHS pulling out all removed items
            if (rhs ~= nil) then
                for _, value in pairs(rhs) do
                    local guid = getValueGuid(value);
                    local lhsByGuid = getValueByGuid(lhs, guid);
                    if (lhsByGuid == nil) then
                        walkTree(nil, value, guid, thisNode);
                    end
                end
            end

        else

            print (" ");
            print ("***********************************************");
            print ("******************  WARNING  ******************");
            print ("***********************************************");
            print ("** Unhandled indexed array found:"..thisNode.name);
            print ("** Data NOT copied!                          **");
            print ("** Data NOT copied!                          **");
            print ("** Data NOT copied!                          **");
            print ("***********************************************");
            print (" ");

--            print ("ignoring"); --  todo: treat with diffpatch?
        end

    elseif (type(lhs) == "userdata") or (type(lhs) == "function" or type(lhs) == "thread") then
        -- userdata, functions, threads ignored

    else
        if (lhs ~= nil and rhs ~= nil) then
            thisNode.action = ((lhs == rhs) and NODE_UNCHANGED) or NODE_CHANGED;
        end
    end
end



local outputTree = {}
function findOrAddOutputNode(node)
    if (node.outputNode ~= nil) then
        --  already exists, so don't add it again.
        return node.outputNode;
    end 

    if (node.parent == nil) then
        -- handle root node.
        node.outputNode = outputTree;
        return outputTree;
    else
        --  doesn't yet exist, add it.
        local parentNode = findOrAddOutputNode(node.parent);
        if (node.action == NODE_REMOVED) then
            if (parentNode.isGuidKeyed or (node.key == nil)) then
                --  removing an indexed array item
                table.insert(parentNode, DELETE_PREFIX..(node.key or node.rhs));

            else
                -- removing a key'ed item
                parentNode[DELETE_PREFIX..(node.key or node.rhs)]= 0;
            end
        else
            if (parentNode.isGuidKeyed or (node.key == nil)) then
                -- indexed table
                if (type(node.lhs) == "table") then
                    -- child table
                    node.outputNode = {}
                    table.insert(parentNode, node.outputNode);
                    return node.outputNode;
                else
                    --  indexed value (not keyed value)
                    table.insert(parentNode, node.lhs);
                end
            else
                -- keyed table.
                if (type(node.lhs) == "table") then
                    -- child table
                    parentNode[node.key] = {};
                else
                    --  value only
                    parentNode[node.key] = node.lhs;
                end
                node.outputNode = parentNode[node.key];
                return parentNode[node.key];
            end
        end
    end
end

function createOutput(node)
    --  iterate the nodes that have changed
    if (node ~= nil) then
        if (node.action ~= nil and node.action ~= NODE_UNCHANGED) then
            findOrAddOutputNode(node);
        end
        if (node.children ~= nil) then
            for _, value in pairs(node.children) do
                createOutput(value);
            end
        end
    end
end

local numChanges = 0;

function printFullTree(node)
    numChanges = 0;
    printTree(node);
    print("Found "..numChanges.." changes")
end

function printTree(node)
    local key = node.key or "?";
    if (node.action ~= nil and node.action ~= NODE_UNCHANGED) then
        numChanges = numChanges + 1;
        if (type(node.lhs) == "table") then
            print (node.name.."=list["..key.."] : ".. node.action);
        elseif (type(node.lhs) == "nil") then
            if (type(node.rhs) == "table") then
                print (node.name.."list["..key.."] : "..node.action);
            else
                print (node.name.."=nil : ".. node.action);
            end
        elseif (type(node.lhs) == "boolean") then
            print (node.name.."="..(node.lhs and "true" or "false").." : ".. node.action);
        else
            print (node.name.."="..node.lhs.. " : ".. node.action);
        end
    end
    if (node.children ~= nil) then
        for _, value in pairs(node.children) do
            printTree(value);
        end
    end
end


function diffPatch:setGuidLookupEnabled(isEnabled)
    INCLUDE_GUID_LOOKUPS = isEnabled;
end

function diffPatch:addGuidField(fieldName)
    table.insert(GUID_FIELDS, fieldName)
end 

--LHS - the new lua tree
--RHS - the old lua tree to compare to
function diffPatch:createDiff(lhs, rhs)
    local rootNode = {name = "."};
    --  walk through the entire source trees looking for changes
    walkTree(lhs, rhs, "root", rootNode);

--    printFullTree(rootNode);    --  for debugging

    --  generate the outputs showing only the valid changes.
    outputTree = {} --  reset the output tree
    createOutput(rootNode)

    return outputTree.root;
end


return diffPatch
