local dialogManager = {}

local List = require(cc.PACKAGE_NAME .. ".libraries.list")

function dialogManager.cleanup()
    echoInfo("dialogManager.cleanupManager()")
    dialogManager.currentDialogStack = List:new()
    dialogManager.currentDialogBreaker = nil
    dialogManager.currentParams = nil
    dialogManager.isShowingDialog = false
    dialogManager.lastDialogTime = 0
    dialogManager.isWaitingForScheduledThread = false
    dialogManager.isQuitOnStack = false
    dialogManager.gameDialogManager = nil
end

function dialogManager.getDialogToDisplay()
    return dialogManager.currentDialogStack:peekLeft()
end

function dialogManager.getCallbackFunction()
    local dialogParameters = dialogManager.getDialogToDisplay()
    if dialogParameters then
        return dialogParameters[5]
    else
        return nil
    end
end

function dialogManager.getStackCount()
    return dialogManager.currentDialogStack:count()
end

function dialogManager.displayNextDialog()
    -- echoInfo("dialogManager.displayNextDialog()")
    -- print(dialogManager.getStackCount(),dialogManager.isWaitingForScheduledThread)
    if (dialogManager.getStackCount() > 0) and (not dialogManager.isWaitingForScheduledThread) then
        -- echoInfo("Dialog to show")
        dialogManager.isShowingDialog = true

        dialogManager.currentParams = dialogManager.getDialogToDisplay()
        dump(dialogManager.currentParams,"dialogParameters")
        local timeSinceLastDialog = os.time() - dialogManager.lastDialogTime
        -- don't show more than 1 dialog per second
        if timeSinceLastDialog < 1.0 then
            print("Schedule dialog")
            dialogManager.isWaitingForScheduledThread = true
            scheduler.performWithDelayGlobal(function()
                if dialogManager.currentParams ~= nil then
                    dialogManager.currentDialogBreaker = ui.showSystemDialog(unpack(dialogManager.currentParams))
                end
                dialogManager.isWaitingForScheduledThread = false
            end, 1 - timeSinceLastDialog)
        else 
            print("show it now")
            dialogManager.currentDialogBreaker = ui.showSystemDialog(unpack(dialogManager.currentParams))
        end
    else
        echoInfo("No dialogs available right now")
    end
end

function dialogManager.popDialog()
    echoInfo("dialogManager.popDialog")
    dialogManager.lastDialogTime = os.time()
    if dialogManager.getStackCount() > 0 then
        return dialogManager.currentDialogStack:popLeft()
    else
        return nil
    end
end

function dialogManager.pushDialog(dialogInfo)
    echoInfo("dialogManager.pushDialog")
    dialogManager.currentDialogStack:pushRight(dialogInfo)
    if not dialogManager.isShowingDialog then
        dialogManager.displayNextDialog()
    end
end

function dialogManager.onChangeScene(event)
    echoInfo("dialogManager.onChangeScene")
    dialogManager.displayNextDialog()
end

function dialogManager.willChangeScene(event)
    echoInfo("dialogManager.willChangeScene")
    if dialogManager.isPlatformDialogDisplayed() then
        -- echoInfo("Currently showing dialog")
        -- dump(dialogManager.currentParams)
        dialogManager.currentDialogStack:pushLeft(dialogManager.currentParams)
    end

    dialogManager.isShowingDialog = false
    dialogManager.isWaitingForScheduledThread = false
end

-- back button functions

function dialogManager.closeAppDialog()
    if ANDROID_INSTANT_CLOSE then
        game.exit()
        return
    end

    local function closeAppButton(button)
        dialogManager.isQuitOnStack = false
        if button == 'OK' then
            gameManager.savePlatformSettings()
            local function onCompleted()
                game.exit()
            end
            if gameManager.getIsInGame() then
                persistence.force(nil, nil, true, onCompleted)
            else
                onCompleted()
            end
        end
    end

    if dialogManager.isQuitOnStack then return end
    dialogManager.isQuitOnStack = true
    dialogManager.showDialog(
            SKELETON_MESSAGES.quitGame.Title, 
            SKELETON_MESSAGES.quitGame.Desc, 
            false,
            {
                ui.SYSTEM_DIALOG_BUTTON_CANCEL,
                ui.SYSTEM_DIALOG_BUTTON_OK
            }, 
            closeAppButton
        )
    return true
end

function dialogManager.onBackKeyPress()
    if dialogManager.isShowingDialog then
        if type(dialogManager.currentDialogBreaker) == "function" then
            local tempDialogBreaker = dialogManager.currentDialogBreaker
            dialogManager.currentDialogBreaker = nil
            tempDialogBreaker("BACK")
            return true
        end
    end
    return false
end


-- public functions

function dialogManager.setGameDialogManager(gameDM)
    dialogManager.gameDialogManager = gameDM
end

function dialogManager.isDialogDisplayed()
    if (dialogManager.gameDialogManager == nil) then return dialogManager.isShowingDialog end
    return dialogManager.isShowingDialog or dialogManager.gameDialogManager.isDialogDisplayed()
end

function dialogManager.isPlatformDialogDisplayed()
    return dialogManager.isShowingDialog
end


function dialogManager.showDialog(title, description, isError, buttonsList, callback, args)
    echoInfo("dialogManager.showDialog()")
    if (title == nil) or (description == nil) then
        echoInfo("********************")
        echoInfo("* WARNING:")
        echoError("* dialogManager.showDialog() has null Parameter (%s)", ((title == nil) and "title" or "description"))
        echoInfo("Title=%q", title or "UNDEFINED")
        echoInfo("Description=%q", description or "UNDEFINED")
        echoInfo("********************")
        return
    end

    echoInfo("dialogManager.showDialog(%q, %q)", title, description)

    dialogManager.pushDialog({title, description, isError, buttonsList, callback, args})
end

function dialogManager.removeCurrentDialog()
    echoError("dialogManager.removeCurrentDialog")
    if dialogManager.isShowingDialog then
        echoInfo("dialogManager.isShowingDialog")
        dialogManager.isShowingDialog = false
        local result = dialogManager.getCallbackFunction()
        dialogManager.popDialog()
        dialogManager.displayNextDialog()
        if result then
            return result
        end
    end
end

function dialogManager.description()
    local currentCallbackFunction = dialogManager.getCallbackFunction()
    if not currentCallbackFunction then
        currentCallbackFunction = "no callback"
    end
    return "currentCallbackFunction: \n"..tostring(currentCallbackFunction)..'\ncurrentDialogStack: \n'..dialogManager.currentDialogStack:description()
end


-- public callback functions

function dialogManager.activateCallback(button)
    local currentCallbackFunction = dialogManager.getCallbackFunction()
    local result = false
    if type(currentCallbackFunction) == "function" then
        currentCallbackFunction(button)
        result = true
    end
    currentCallbackFunction = nil
    return result
end


-- initialisation

function dialogManager.init()
    dialogManager.cleanup()
    callbackManager.listenForCallback(EVENT_WILL_CHANGE_SCENE,function(event) dialogManager.willChangeScene(event) end)
    backButtonManager.registerFunctionWithPriority(dialogManager.onBackKeyPress, 10, "systemDialog")
    backButtonManager.registerFunctionWithPriority(dialogManager.closeAppDialog, 0, "quitGame")
end

dialogManager.init()

-- specific dialogs

function dialogManager.showGDPRDialog(callback)
    echoInfo("Show GDPR dialog - about to check if necessary")
    if not SHOW_GDPR_DIALOG then return true end
    if persistenceUserGame then
        if persistenceUserGame.gdpr_acceptance then
            echoInfo("GDPR already accepted")
            return true
        end
    end

    dialogManager.showDialog( 
        SKELETON_MESSAGES.GDPR.Title, 
        SKELETON_MESSAGES.GDPR.Desc,
        false,
        {ui.SYSTEM_DIALOG_BUTTON_LONG}, 
        callback,
        {
            ignoreConfigUIBuild = true,
            GDPR = true,
            -- Don't allow dialog to be skipped with android back button
            blockBackButton = true
        }
    )
end


return dialogManager