-- dlcManager.lua
local M = {}
--[[
Flow pseudocode:

    startSync called to start process
    {
        download full JSON list
        var outstandingJSON = determine all missing/out of date JSON file
        if (outstandingJSON != empty)
        {
            callback DLC_ON_DOWNLOAD_START(fileCount, blocking=true)
            foreach (outstandingJSON)
            {
                download file with callback DLC_ON_DOWNLOAD_PROGRESS_UPDATE(bytes, totalBytes, blocking=false)
                callback DLC_ON_DOWNLOAD_PROGRESS_COMPLETEFILE(fileNumber, fileCount, blocking=true)
            }
            callback DLC_ON_DOWNLOAD_COMPLETE_SUCCESS(totalFiles, blocking=true)
        }

        async
        {
            download full resource file list.
            var outstandingFile = all missing files - sorted by plists first, then priority flag 
            
            if (outstandingFile != empty)
            {
                callback DLC_ON_DOWNLOAD_START(fileCount, blocking=false)
                foreach (outstandingFile)
                {
                    download file with callback DLC_ON_DOWNLOAD_PROGRESS_UPDATE(bytes, totalBytes, blocking=false)
                    callback DLC_ON_DOWNLOAD_PROGRESS_COMPLETEFILE(file, blocking=false)
                }
                callback DLC_ON_DOWNLOAD_COMPLETE_SUCCESS(totalFiles, blocking=false)
            }
        }
    }
]]--

----------------------------------------------------
-- Internal Constants
----------------------------------------------------
require "lfs";
local DLC_SERVER_URL = "https://s3.amazonaws.com/%s/dlc/"; 
local DLC_VERSION = 0.0
----------------------------------------------------
-- Public Constants
----------------------------------------------------

-- subscribe to these callback events with dlcManager:subscribeToEvent(event, callback)
M.DLC_ON_DOWNLOAD_START                 = "dlcOnDownloadStart";                  --  onDownloadStart(fileName, isBlocking)
M.DLC_ON_DOWNLOAD_COMPLETE_SUCCESS      = "dlcOnDownloadCompleteSuccess";        --  onDownloadCompleteSuccess(isBlocking)
M.DLC_ON_DOWNLOAD_COMPLETE_ERROR        = "dlcOnDownloadCompleteError";          --  onDownloadCompleteError(errorMessage, isBlocking)
M.DLC_ON_DOWNLOAD_PROGRESS_COMPLETEFILE = "dlcOnDownloadProgressCompleteFile";   --  onDownloadProgressCompleteFile(file, isBlocking)
M.DLC_ON_DOWNLOAD_PROGRESS_UPDATE       = "dlcOnDownloadProgressUpdate";         --  onDownloadProgressUpdate(downloadedBytes, totalBytes, isBlocking)
 
local SALT_LENGTH = 16
local CIPHER = 'aes-128-cfb'

----------------------------------------------------
-- Internal properties
----------------------------------------------------

local StateContainerFactory = require(cc.PACKAGE_NAME .. ".api.StateContainer")
local dlcState = StateContainerFactory.createContainer()
local _eventDispatcher = cc.Director:getInstance():getEventDispatcher();
----------------------------------------------------
-- Internal methods
----------------------------------------------------

-- set this to 1, 2, or 4 to force the download of that dlc (must be on first run)
-- keep it nil otherwise
M.DLC_SCALE_OVERRIDE = nil


function M.raiseEventWithData(eventName, data)
    if eventName == dlcManager.DLC_ON_DOWNLOAD_COMPLETE_SUCCESS and dialogManager.isDialogDisplayed() then
        echoInfo("Error so restarting")
        gameManager.reset()
    else
        local event = cc.EventCustom:new(eventName)
        local jsonData = json.encode(data);
    --    print("Raising event "..eventName.." with data:"..jsonData);
        event._usedata = jsonData;
        _eventDispatcher:dispatchEvent(event)
    end
end

function M.onDownloadStart(fileName, totalDownloads, remainingDownloads, isBlocking)
    M.raiseEventWithData(M.DLC_ON_DOWNLOAD_START,  
        {fileName = fileName, 
        totalDownloads = totalDownloads,
        remainingDownloads = remainingDownloads,
        isBlocking = isBlocking});
end

function M.onDownloadCompleteSuccess(isBlocking)
    if isBlocking then
        M.runAllDlcLuaHacks()
    end
            
    M.raiseEventWithData(M.DLC_ON_DOWNLOAD_COMPLETE_SUCCESS, {isBlocking = isBlocking, successes = (M.successes or 0), failures = (M.failures or 0)});
    --  reset the counters
    M.successes = 0;
    M.failures = 0;
end;


function M.onDownloadProgressUpdate(result, downloadedBytes, totalBytes, isBlocking)
    if result.code == HTTPREQUEST_SUCCESS_NOT_MODIFIED then
        -- ignore feedback if no modifications
    elseif result.code == HTTPREQUEST_SUCCESS then
        M.raiseEventWithData(M.DLC_ON_DOWNLOAD_PROGRESS_UPDATE, 
            { result = result, downloadedBytes = downloadedBytes, totalBytes = totalBytes, isBlocking = isBlocking});
    else
        M.onDownloadCompleteError(result.message, isBlocking);
    end
end;

function M.onDownloadCompleteError(errorMessage, isBlocking)
    M.raiseEventWithData(M.DLC_ON_DOWNLOAD_COMPLETE_ERROR, 
        { result = result, isBlocking = isBlocking, errorMessage = errorMessage});
end;

function M.saveToFile(fileName, data)
    local outfile = io.open(fileName, "w+" )
    outfile:write(data)
    outfile:close()
end

function M:onDownloadProgressCompleteFile(result, originalRequest, file, isBlocking)
    local hasFailed = false;
    if result.code == HTTPREQUEST_SUCCESS then
        local outputFile = originalRequest:getOutputFile()
        local expectedMd5 = nil;
        local md5 = nil
        dump(result, "received File")

        local fileContent = io.readfile(outputFile)
        local decryptedData = M.checkAndDecryptRaw(fileContent)
        md5 = crypto.md5(decryptedData)
        expectedMd5 = file["md5"..M.getFileScaleFactor(file)];

        if (md5 == expectedMd5) then
            echoInfo("Success m5d:%s", md5)

            if ((file.type == 'zip') or (file.type == "zipB")) and not M.unzipFile(file) then
                print ("failed extracting " .. file.file)
                hasFailed = true;
            else
                if (file.res) then
                    self._pendingResourceDownloadCount = self._pendingResourceDownloadCount - 1;
                else
                    if DEBUG_DECODE_LUAHACKS and (file.type == "lua") then
                        M.saveToFile(outputFile, decryptedData)
                    end

                    self._pendingJsonDownloadCount = self._pendingJsonDownloadCount - 1;
                end;
            end
        else
            sendLogToServer({"DLC ERROR:", file.file})
            print("File "..file.file.." downloaded with incorrect MD5")
            print("Expected:"..expectedMd5)
            print("Recieved:"..md5)
            hasFailed = true;
        end;

    else
        print ("failed downloading" .. file.file)
        echoError("Error:%s",result.message)
        hasFailed = true;
    end

    if hasFailed then
        --  if the file failed, then re-add it to the list to re-download
        self:addFileToQueue(file);
        M.failures = (M.failures or 0) + 1
    else
        M.successes = (M.successes or 0) + 1
    end

    if M.failures > 5 then
        dialogManager.showDialog( 
                    SKELETON_MESSAGES.DlcUpdateFailed.Title, 
                    SKELETON_MESSAGES.DlcUpdateFailed.Desc,
                    true,
                    { ui.SYSTEM_DIALOG_BUTTON_OK }, 
                    gameManager.reset,
                    {hsUrl = SKELETON_MESSAGES.DlcUpdateFailed.hsUrl});
        return
    end

    --  if this is a blocking file, then call the startAllDownloads.
    --  this will effectively tell the UI if the last Blocking download is completed.
    --  otherwise, just continue with the next file.
    if (isBlocking) then
        self:startAllDownloads();
    else
        self:startNextDownload();
    end
end

----------------------------------------------------------------------------------------------------------------------------------------



function M.getMasterDlcServerUrl()
    return string.format(DLC_SERVER_URL.."%s/_manifest.json", DLC_VERSION);
end;


--if forFolder param is true then
--  this will return "1x" instead of blank
--  and not have a "-" prefix (as in "-4x") 
--  but will have a trailing "/"
function M.getFileScaleFactor(file, forFolder)
    local scaleSuffix = ""
    if (file.res) and (file["md5-2x"] ~= nil) then
        scaleSuffix = forFolder and "1x" or ""
        --  this is a resource file
        -- see if there is an appropriate scaled version
        if (M.contentScaleFactor ~= "1x") then
            if (file["md5-"..M.contentScaleFactor] ~= nil) then
                scaleSuffix = (forFolder and "" or "-")..M.contentScaleFactor
            end
        end
    end
    if (forFolder) then
        scaleSuffix = scaleSuffix .. "/"
    end
    --echoInfo("Scale suffix for %s is \"%s\"", file.file, scaleSuffix)
    return scaleSuffix;
end;


function M.getDlcServerObjectUrl(file)
    local fullPath = DLC_SERVER_URL..DLC_VERSION.."/"
    fullPath = fullPath..file.file..M.getFileScaleFactor(file).."."..file.type;
    echoInfo("Getting DLC =>\n  %s", fullPath)
    return fullPath;
end;


function M.getDlcLocalObjectPath(file)
    local path = file.localpath;
    local fileName = file.file;
    local fileType = file.type;
    local scaleSuffix = "" --M.getFileScaleFactor(file) -- NOTE: not extracting to res folder
    --local fullPath = device.writablePath; -- old method, stores in documents folder?
    local fullPath = runaway.FileHelpers:getDlcPath();
    if (path) then
        fullPath = fullPath .. path .. "/";
        device.ensureFolderExists(fullPath)
    end;
    fullPath = fullPath .. fileName;
    if (fileType) then
        fullPath = fullPath .. scaleSuffix .."." .. fileType;
    end;
    return fullPath;
end;

--[[ raw content format is:
:16digitsalt:encodedData
]]-- 
local dividerPosition = SALT_LENGTH + 2
function M.isEncryptedContent(rawContent)
    if (rawContent) then
        if (string.sub(rawContent, 1, 1) == ":") then
            if (string.sub(rawContent, dividerPosition, dividerPosition) == ":") then
                return true
            end
        end
    end
    return false
end

function M.splitSaltAndData(rawSaltAndData)
    --dump(debug.traceback(), "splitSaltAndData")
    if M.isEncryptedContent(rawSaltAndData) then
        local salt = string.sub(rawSaltAndData, 2, SALT_LENGTH + 1)
        local rawData = string.sub(rawSaltAndData, SALT_LENGTH + 3, string.len(rawSaltAndData))
        return salt, rawData
    else
        return nil, rawSaltAndData
    end
end


function M.decryptRaw(rawSaltAndData)
    local key = runaway.GameConfig:getAESJSONKey()
    local salt, rawData = M.splitSaltAndData(rawSaltAndData)
    -- local iv = salt or runaway.GameConfig:getAESJSONIV()
    local text = crypto.decodeBase64(rawData)
    local decryptedContent = crypto.unpad(crypto.decrypt(CIPHER, text, key, salt))

    -- if DEBUG == 2 then
    --     echoInfo("Decrypting data with key:%q  salt:%q", key, salt)
    --     dump(decryptedContent, "decrypted result")
    -- end
    return decryptedContent
end

function M.checkAndDecryptRaw(rawData)
    if (rawData ~= nil) and M.isEncryptedContent(rawData) then
        return M.decryptRaw(rawData)
    else
        return rawData
    end
end

--  get the file contents, and decrypt them if required
function M.getFileContents(path)
    local fileContents = io.readfile(path);
    return M.checkAndDecryptRaw(fileContents)
end

function M.getFileHash(path)
    --assumes that the file has been previously checked as existing
    local hash = nil;
    local fileContents = M.getFileContents(path);
    if fileContents ~= nil then
        hash = crypto.md5(fileContents);
    end
    return hash;
end

function M.isDownloadRequired(file, ignoreMd5)
    --print("searching for ".. file.fullPath)
    local required = true

    if (file.localpath == "android/") and (device.client_platform ~= "android") then 
        echoInfo("Skipping android DLC file: %s", file.file)
        required = false
    elseif (file.localpath == "ios/") and (device.client_platform ~= "ios") then 
        echoInfo("Skipping ios DLC file: %s", file.file)
        required = false
    end

    if required then
        --  if the file already exists, then check if it hasn't changed
        if (cc.FileUtils:getInstance():isFileExist(file.fullPath)) then
            local existingHash = M.getFileHash(file.fullPath);
            local scaleFactor = M.getFileScaleFactor(file, false)
            local newHash = file["md5" .. scaleFactor] or file.md5;
            echoInfo("checking if %s needs to be downloaded old:%s <=> new:%s", file.file, existingHash, newHash)
            required = (existingHash ~= newHash) or ignoreMd5
        end
    end
    return required
end;


function M.decryptJson(rawData)
    return json.decode(M.checkAndDecryptRaw(rawData))
end

----------------------------------------------------------------------------------------------------------------------------------------
function M.unzipFile(file)
    print ("Unzipping file "..file.fullPath);
    --local dlcPath = device.writablePath;
    local dlcPath = runaway.FileHelpers:getDlcPath()

    --  if this is a resource file, then extract it to the appropriate 1x, 2x, 3x, 4x folder
    -- NOTE: Not doing this anymore
    -- if (file.res) then
    --     dlcPath = dlcPath .. M.getFileScaleFactor(file, true);
    -- end
    device.ensureFolderExists(dlcPath);

    local results = runaway.FileHelpers:extractZipFile(file.fullPath, dlcPath);
    local resultData = json.decode(results);
    dump(results, "results")
    dump(resultData, "resultData")
    if (resultData.success) then
        print ("Unzipping file "..file.fullPath..":\n successful");
    else
        print ("Unzipping file "..file.fullPath..":\n failed");
    end
    print ("  into folder:\n   ".. dlcPath)
    return resultData.success;
end


function M:startAllDownloads()
    if table.isEmpty(self._pendingJsonDownloads) then
        --  inform the UI that all blocking downloads are completed
        if not M.requireNonBlockingDlc or table.isEmpty(self._pendingResourceDownloads) then
            M.onDownloadCompleteSuccess(true)
        end
    end
    --  continue with the downloads
    self:startNextDownload()
end

function M:reExtractAllZips()
    -- iterate through all of the zip files and re-extract them.
    print ("Re-extracting the zip files:")
    --dump(M.checkAndDecryptRaw(rawData), "raw DLC Manifest")
    local data = M.decryptJson(self._rawMasterData)

    if data then
        for i, file in pairs(data) do
            file.fullPath = M.getDlcLocalObjectPath(file);
            if ((file.type == 'zip') or (file.type == "zipB")) then
                if M.isDownloadRequired(file, true) then
                    if (cc.FileUtils:getInstance():isFileExist(file.fullPath)) then
                        echoInfo("Re-extracting %s", file.fullPath)
                        M.unzipFile(file)
                    end
                end
            end
        end
    end    
end

function M:startNextDownload()
    local file = nil
    local isBlocking = false

    -- If a file has finished downloading AND it was a blocking file,
    -- increment count of blocking files downloaded
    if self._currentFileDownload and M:isFileBlocking(self._currentFileDownload) then
        self._completedBlockingDlcCount = self._completedBlockingDlcCount + 1
    end

    if (next(self._pendingJsonDownloads)) then
        file = table.remove(self._pendingJsonDownloads, 1)
        isBlocking = true;
    elseif (next(self._pendingResourceDownloads)) then
        file = table.remove(self._pendingResourceDownloads, 1)
        isBlocking = M.requireNonBlockingDlc
    end

    -- Save ref to current file that is being downloaded. This may be nil if
    -- all files have been downloaded
    self._currentFileDownload = file

    if (file) then
        local isResource = false
        if (file.res) then
            M.onDownloadStart(file.id, self._totalResourceDownloadCount, self._pendingResourceDownloadCount, isBlocking)
            isResource = true
        else
            M.onDownloadStart(file.id, self._totalJsonDownloadCount, self._pendingJsonDownloadCount, isBlocking)
        end

        local reqUrl = M.getDlcServerObjectUrl(file)
        print("Attempting to download "..reqUrl .. (isBlocking and " (Blocking)" or " (non blocking)" ))

        echoInfo("downloading DLC file %s", file.fullPath)
        local req = network.createHTTPFileRequest(
            function(result, downloaded, total) 
                M.onDownloadProgressUpdate(result, downloaded, total, isBlocking)
            end,
            function(result, req) 
                print("Download complete:\n"..reqUrl);
                self:onDownloadProgressCompleteFile(result, req, file, isBlocking)
            end,
            reqUrl, file.fullPath, file.id)
        network.getFile(req)
    else
        --  all done
        print("update completed")
        if M.resetManifest then
            print("Forcing reExtractAllZips")
            M:reExtractAllZips()
        end
        self:markDlcUpdateSuccessful();
        M.onDownloadCompleteSuccess(isBlocking)
    end
end

function M:getDlcCounts()
    local total = (self._totalResourceDownloadCount or 0) + (self._totalJsonDownloadCount or 0)
    local pending = (self._pendingResourceDownloadCount or 0) + (self._pendingJsonDownloadCount or 0)
    local completed = total - pending

    return completed, total
end
--------------------

function M:getBlockingDlcCounts()

    -- Get the total and completed count for blocking dlc files. Blocking dlc
    -- are files that MUST be downloaded before the game will start.

    local total = self._totalJsonDownloadCount or 0
    local completed = self._completedBlockingDlcCount or 0

    return completed, total
end
--------------------

function M:blockingDlcRemains()

    -- Checks if there are blocking dlc files that are queued for download or
    -- are currently being downloaded.

    local blockingDlcRemains = self._completedBlockingDlcCount < self._totalJsonDownloadCount or
                                (self._currentFileDownload and M:isFileBlocking(self._currentFileDownload))
    return blockingDlcRemains
end
--------------------

function M:isFileBlocking(file)

    -- Checks if the given file is blocking dlc. Blocking dlc is dlc that MUST
    -- be downloaded before the game will start.

    local isBlocking = (not file.res or      -- file is not a resource
                        file.type == 'zipB') -- OR file is a zipB (B denotes blocking)
    return isBlocking
end
--------------------

function M:markDlcUpdateSuccessful()
    local etagStore = {}
    etagStore.masterJsonEtag = self._currentMasterJsonEtag;
    etagStore.rawMasterData = self._rawMasterData;
    M.resetManifest = false
    dlcState.save(etagStore);
end;


function M:clearQueue()
    self._pendingJsonDownloads = {};
    self._pendingResourceDownloads = {};
    self._totalJsonDownloadCount = 0;

    -- Number of blockingDlc files downloaded. Blocking dlc files are those
    -- that MUST be downloaded before the game starts
    self._completedBlockingDlcCount = 0;   

    self._pendingJsonDownloadCount = 0;
    self._pendingResourceDownloadCount = 0;
    self._totalResourceDownloadCount = 0;
end;


function M:addFileToQueue(file)
    if (file.res) and (file.type ~= 'zipB') then
        table.insert(self._pendingResourceDownloads, file);
        self._pendingResourceDownloadCount = self._pendingResourceDownloadCount + 1;
        self._totalResourceDownloadCount = self._totalResourceDownloadCount + 1; 
    else
        table.insert(self._pendingJsonDownloads, file);
        self._pendingJsonDownloadCount = self._pendingJsonDownloadCount + 1;
        self._totalJsonDownloadCount = self._totalJsonDownloadCount + 1;
    end;
end;

function M.runAllDlcLuaHacks()
    echoInfo("RUNNING LUA HACKS!")
    local dlcStateValues = nil
    local masterData = nil
    if M._rawMasterData == nil then
        dlcStateValues = dlcState.load()
        if (dlcStateValues and dlcStateValues.rawMasterData) then
            M._rawMasterData = dlcStateValues.rawMasterData
        end
    end
    if (M._rawMasterData) then
        local masterData = M.decryptJson(M._rawMasterData)

        for k, file in pairs(masterData) do
            if file.type == "lua" then
                local filePath = M.getDlcLocalObjectPath(file)
                echoInfo("running hack in file %s.lua", file.file)
                local fileContents = M.getFileContents(filePath)
                assert(loadstring(fileContents))()
            end
        end
    end

    -- local fileAsString = cc.FileUtils:getInstance():getStringFromFile(cc.FileUtils:getInstance():fullPathForFilename("hacks.lua"))
    -- assert(loadstring(fileAsString))()
end

function M.isFolderInManifest(manifest, dlcSubFolder, folderName)
    for fileKey, file in pairs(manifest) do
        local fullFilename = string.format("%s.%s", file.file, file.type)
        if (file.file == folderName) and (file.type == "zip" or file.type == "zipB") then
            --echoInfo("found zip folder in manifest: %s/%s", dlcSubFolder, folderName)
            return true, false --   found,but DONT-recurse
        elseif (file.localpath) and (string.len(file.localpath) ~= 0) then
            local localpath = file.localpath
            --  strip trailing slash off localpath 
            if string.sub(localpath, -1) == "/" then localpath = string.sub(localpath, 1, -2) end
            local relativepath = folderName
            if string.len(dlcSubFolder) ~= 0 then
                relativepath = dlcSubFolder .. "/" .. folderName
            end
            if relativepath == localpath then
                --echoInfo("found localpath folder in manifest: %s/%s", dlcSubFolder, folderName)
                return true, true ----   found,DO-recurse
            end
        end
    end
    echoInfo("unable to find folder in manifest: %s/%s", dlcSubFolder, folderName)
    return false    --notfound
end

function M.isFileInManifest(manifest, dlcSubFolder, fileName)
    for fileKey, file in pairs(manifest) do
        local fullFilename = string.format("%s.%s", file.file, file.type)
        local localpath = file.localpath
        --  strip trailing slash off localpath 
        if string.sub(localpath, -1) == "/" then localpath = string.sub(localpath, 1, -2) end

        if (fullFilename == fileName) and (localpath == dlcSubFolder) then
            --echoInfo("found file in manifest: %s/%s", dlcSubFolder, fileName)
            return true --ie:don't delete
        end
    end
    echoInfo("unable to find file in manifest: %s/%s", dlcSubFolder, fileName)
    return false -- nerfed
end

function M.triggerResetManifest()
    print ("triggerResetManifest")
    M.resetManifest = true
    dlcState.delete()
end

--  clears out any old files/folders not referenced in the manifest
function M.cleanupDlcFolder(files, dlcSubFolder)
    require "lfs"
    local dlcPath = runaway.FileHelpers:getDlcPath();
    local dlcFullPath = dlcPath
    if dlcSubFolder and (string.len(dlcSubFolder) ~= 0) then
        dlcFullPath = dlcPath .. dlcSubFolder .. "/"
    else
        dlcSubFolder = ""   --  ensure it's a string
    end 
    local exists = cc.FileUtils:getInstance():isFileExist(dlcFullPath)
    if exists then
        echoInfo("Cleaning up DLC folder : %s", dlcFullPath)
        for file in lfs.dir(dlcFullPath) do
            local fullPath = dlcFullPath .. file
            local relativepath = file
            if string.len(dlcSubFolder) ~= 0 then
                relativepath = dlcSubFolder .. "/" .. file
            end
            local mode = lfs.attributes(fullPath, "mode")
            if mode then
                if mode == "directory" then
                    --  it's a directory
                    --  the "luaboot" folder is a special case for the firstLoader
                    if (file ~= ".") and (file ~= "..") and (file ~= "luaboot") then
                        local found, recurse = M.isFolderInManifest(files, dlcSubFolder, file) 
                        if not found then
                            device.removePath(dlcPath..relativepath)
                        elseif recurse then
                            --  recurse into folder
                            M.cleanupDlcFolder(files, relativepath)
                        end
                    end
                else
                    --  it's a file
                    if not M.isFileInManifest(files, dlcSubFolder, file) then
                        device.removePath(dlcPath..relativepath)
                    end
                end
            end
        end
    end
end

function M:onDlcManifestDownloaded(result, rawData, originalRequest)
    print("onDlcManifestDownloaded")
    if (result.code == HTTPREQUEST_SUCCESS_NOT_MODIFIED) then
        -- the server file hasn't been modified since our last successful update, so just return success.
        print ("DLC Master not changed. No updates required")
        M.onDownloadCompleteSuccess(true)
    elseif (result.code == HTTPREQUEST_SUCCESS) then
        --dump(result.headers, "result.headers");
        if result.headers["Etag"] ~= nil then
            self._currentMasterJsonEtag = result.headers["Etag"]
        else
            self._currentMasterJsonEtag = result.headers["ETag"]
        end

        self._rawMasterData = rawData;

        print ("Received DLC Manifest with ETag:"..(self._currentMasterJsonEtag or ""));
 
        --dump(M.checkAndDecryptRaw(rawData), "raw DLC Manifest")
        local data = M.decryptJson(rawData);
        self:clearQueue();

        if (data == nil) then
            M.onDownloadCompleteError("Unable to decypher updates list from server.", true);
        else
            for i, file in pairs(data) do
                file.fullPath = M.getDlcLocalObjectPath(file);
                if (M.forceDownload or M.isDownloadRequired(file))then
                    self:addFileToQueue(file);
                else
                    print (file.file .. " is up to date.. skipping")
                end
            end;

            if DLC_DONT_CLEANUP == nil or (DLC_DONT_CLEANUP ~= nil and not DLC_DONT_CLEANUP) then
                M.cleanupDlcFolder(data)
            end

            self:startAllDownloads();
        end
    else
        --  todo: implement correctly
        M.onDownloadCompleteError("Unable to download updates list from server.", true);
    end; 
end;



----------------------------------------------------
-- Public methods
----------------------------------------------------


--  init.  Called automatically below
function M:init()
    print("dlcManager Constructor")

    local function dlcStateListener(event)
        echoInfo("dlcStateListener name:%s values:%s encode:%s time:%s", event.name, event.values, event.encode, event.time)
        return event.values
    end

    --  make sure that we have the DLC folder created
    runaway.FileHelpers:getDlcPath();
    -- preprime the scaleFactor value
    M.calculateContentScaleFactor();
    --  add a search path to the DLC folder
    local dlcPath = runaway.FileHelpers:getDlcPath();
    local searchPaths = cc.FileUtils:getInstance():getSearchPaths()
    if (not table.contains(searchPaths, dlcPath)) then
        cc.FileUtils:getInstance():addSearchPath(dlcPath, true)
    end


    dlcState.init(dlcStateListener, "dlc", "dlcManagerKey")
end;

function M:setServer(dlcServerUrl, dlcVersion)
    DLC_SERVER_URL = dlcServerUrl
    DLC_VERSION = dlcVersion or "0"
    echoInfo("Setting DLC server to %q, version %s", DLC_SERVER_URL, DLC_VERSION)
end

function M:reset()
    dlcState.delete();
    dlcState.save({});
    device.removePath(runaway.FileHelpers:getDlcPath());
end;


--  startSync.  start the sync process
function M:startSync(forceDownload, requireNonBlockingDlc)
    -- forceDownload = true--   testing -force all to download
    -- requireNonBlockingDlc = true--   testing -force all to download
    --  reset the counters
    M.successes = 0;
    M.failures = 0;

    echoInfo("Starting DLC Sync with requireNonBlockingDlc=%s", (requireNonBlockingDlc and "true" or "false"))
    self:clearQueue()

    -- get the master list of JSON files
    local reqUrl = M.getMasterDlcServerUrl();
    -- "https://s3.amazonaws.com/flutterstaticfilesdev/dlc/1.9.json"
    print("Master DLC Url =" .. reqUrl);
    local reqTag = "DLC Master JSON";

    M.forceDownload = forceDownload or false
    M.requireNonBlockingDlc = requireNonBlockingDlc or false
    --  calculate the required content scale (which also primes the internal variable)
    M.calculateContentScaleFactor() 

    M.onDownloadStart(fileName, 1, 1, false);

    local req = network.createHTTPRequest(
        function(result, jsonData, originalRequest)     
            self:onDlcManifestDownloaded(result, jsonData, originalRequest) 
        end,
        reqUrl, HTTPREQUEST_GET, reqTag);

    --  add etag query
    local dlcStateValues = dlcState.load();
    if (dlcStateValues and dlcStateValues.masterJsonEtag) then
        if (not M.forceDownload and not M.resetManifest) then
            local etagQuery = [[If-None-Match: "]]..dlcStateValues.masterJsonEtag..[["]];
            print("Adding additional header\n"..etagQuery);
            req:setHeaders({etagQuery});
        end
    -- If ALL DLC is required, reextractAllZips when download is complete, 
    -- otherwise don't. requireNonBlockingDlc is false the first the game is 
    -- run.
    elseif requireNonBlockingDlc then
        -- Extraction of each zip file occurs after each file has downloaded, 
        -- this trigger causes a reextraction of ALL zips.
        M.resetManifest = true
    end;

    network.send(req)
end;

function M:isDownloadInProgress()
    return (self._pendingJsonDownloadCount + self._pendingResourceDownloadCount) > 0
end

function M.getEncryptedDlc(fileId)
    local function parseFile(path)
        local fileString = cc.FileUtils:getInstance():getStringFromFile(path)
        if fileString ~= nil then
            return fileString
        else
            echoInfo("file not found")
        end
    end

    local dlcStateValues = dlcState.load();
    if (dlcStateValues and dlcStateValues.rawMasterData) then
        local masterData = M.decryptJson(dlcStateValues.rawMasterData)
        local fileDetails = masterData[fileId]
        dump(masterData, "found Files")
        if fileDetails then
            local filePath = M.getDlcLocalObjectPath(fileDetails)
            return M.getFileContents(filePath)
        end
    end
    return nil
end




M.contentScaleFactor = nil;
M.contentScaleFactorInt = 1
function M.calculateContentScaleFactor()
    if (M.contentScaleFactor) then
        return M.contentScaleFactor;
    else
        local scaleFactor = tonumber(display.contentScaleFactor)

        if M.DLC_SCALE_OVERRIDE then
            scaleFactor = M.DLC_SCALE_OVERRIDE
        end

        local scale = string.format("%dx", scaleFactor)
        M.contentScaleFactorInt = scaleFactor;

        M.contentScaleFactor = scale;
        echoInfo("Cocos claims that this device should use "..scaleFactor.."x scaled resources")
        echoInfo("Determined device scale factor = ".. scale)

        --  Also set the scaleFactor resource search order while we're here.
        local resolutionSearchOrder = {"1x"}
        if (scale == "4x") then
            resolutionSearchOrder = {"4x","3x","2x","1x"}
        elseif (scale == "3x") then
            resolutionSearchOrder = {"3x","2x","1x"}
        elseif (scale == "2x") then
            resolutionSearchOrder = {"2x","1x"}
        end
        cc.FileUtils:getInstance():setSearchResolutionsOrder(resolutionSearchOrder)

        return scale;
    end
end

--  determines what size scaling 
function M.determineResourceScaleFromFile(file)
    local fullFilename=cc.FileUtils:getInstance():fullPathForFilename(file);
    local itemScale = 1;
    if (string.find(fullFilename, "/4x/")) then
        itemScale = 4
    elseif (string.find(fullFilename, "/3x/")) then
        itemScale = 3
    elseif (string.find(fullFilename, "/2x/")) then
        itemScale = 2
    end

    return itemScale, (M.contentScaleFactorInt / itemScale);  --  item scale, item scale multiplier
end


-- initialise before we return the manager
M:init();

return M;
