local gameManager = class("gameManager")

local StateContainerFactory = require(cc.PACKAGE_NAME .. ".api.StateContainer")

local Server        = require(cc.PACKAGE_NAME..".server.server")
local OAuth2        = require(cc.PACKAGE_NAME..".server.OAuth2")
local scheduler     = require(cc.PACKAGE_NAME..".scheduler")

local serverManager = require(cc.PACKAGE_NAME..".managers.serverManager")
local authManager   = require(cc.PACKAGE_NAME..".server.lcdFreeManager")

--Globals
eventDispatcher     = cc.Director:getInstance():getEventDispatcher()
dlcManager          = require(cc.PACKAGE_NAME..".managers.dlcManager")
persistence         = require(cc.PACKAGE_NAME..".persistence")
crashManager        = require(cc.PACKAGE_NAME..".managers.crashManager")
socialManager       = require(cc.PACKAGE_NAME..".managers.socialManager")
timeManager         = require(cc.PACKAGE_NAME..".managers.timeManager")
publicCounterManager= require(cc.PACKAGE_NAME..".managers.publicCounterManager")
audioManager        = require(cc.PACKAGE_NAME..".managers.audioManager")
utils               = require(cc.PACKAGE_NAME..".utils")
metadataManager     = require(cc.PACKAGE_NAME..'.managers.metadataManager')
platformUserGameData= require(cc.PACKAGE_NAME..'.userGameData')
activityFeedManager = require(cc.PACKAGE_NAME..'.managers.activityFeedManager')

require(cc.PACKAGE_NAME .. ".managers.gameManager-conditions")

userAccount         = {}
friends             = {}

gameManager.devInfo = {}

gameManager.currentScene = "";
gameManager.SCENE_LOADING     = "loadingAppScene"
gameManager.SCENE_START       = "startScene"
gameManager.SCENE_GAME        = "loadingGameScene"

gameManager.isNewAccount = false 
gameManager.firstRun = true

gameManager.shouldFireGPUAnalytic = false;

gameManager.gpuTestDummyFilename = "GPUTest"
gameManager.dlc_url = nil
gameManager.posts_res_url = nil

-- these are used for login progress % complete
local PROGRESS_START     = 0
local PROGRESS_INIT      = 1
local PROGRESS_SETTINGS  = 2
local PROGRESS_DLC       = 3
local PROGRESS_LOGGED_IN = 4
local PROGRESS_USERDATA  = 5
local PROGRESS_GAME      = 6


function gameManager:ctor()
    print ("gameManager->ctor()")
end

function gameManager:onExit()
    print ("gameManager->onExit()")
end



function gameManager:init(verifyFunc, beforeSaveFunc, getGameDescriptionFunc, getHardCFunc, onResetFunc, getScoreFunc, onUserLoggedInFunc)
    print ("gameManager->init()")
    gameManager.SetLoginProgress(PROGRESS_START)

     --  initialise the server managers etc
    local server = nil
    if LOCAL_MODE then
        server = Server.new()
    else
        server = OAuth2.new(gameManager.devInfo["key"], gameManager.devInfo["secret"], params)
    end

    server:setEndPoints(endpoints)
    network.enableCookies(COOKIE_FILE)
    network.setTimeoutForRead(15)       --  default timeout is 1 minute, reduce to 15 secs

    echoInfo("gameManager.Init, creating serverManager with OAuth server:")

    platformUserGameData.init()
    timeManager.init()

    local headers = {}
    headers["device-id"] = device.getDeviceIdentifier()
    
    serverManager.init(headers, server, SERVER_ENDPOINTS)

    --  switch to the loadingApp scene to start with
    --  so that we actually have a scene (with a SYSTEMLAYER to draw error dialogs onto)
    gameManager.switchScene(gameManager.SCENE_LOADING)

    crashManager.init()
    dlcManager.init()

    gameManager.verifyFunc = verifyFunc
    gameManager.getScoreFunc = getScoreFunc or function() return 0 end
    gameManager.onUserLoggedInFunc = onUserLoggedInFunc
    gameManager.getHardCFunc = getHardCFunc or function() return 0 end

    gameManager.authManager = authManager
    if not LOCAL_MODE then
        gameManager.authManager:init(gameManager.devInfo)
        gameManager.authManager.setOnNewSessionFunc(gameManager.onNewLcdSession)
    end

    gameManager.authState     = StateContainerFactory.createContainer()
    gameManager.platformState = StateContainerFactory.createContainer()

    gameManager.authState.init(nil, "authentication", "key", DEBUG)
    gameManager.authState.load()

    gameManager.platformState.init(nil, "platform", "key", DEBUG)
    gameManager.platformData = gameManager.platformState.load() or {}

    --  DLC Events
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(dlcManager.DLC_ON_DOWNLOAD_START                  ,gameManager.onDlcDownloadStart), 1)
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(dlcManager.DLC_ON_DOWNLOAD_COMPLETE_SUCCESS       ,gameManager.onDlcDownloadCompleteSuccess), 1)
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(dlcManager.DLC_ON_DOWNLOAD_COMPLETE_ERROR         ,gameManager.onDlcDownloadCompleteError), 1)
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(dlcManager.DLC_ON_DOWNLOAD_PROGRESS_COMPLETEFILE  ,gameManager.onDlcDownloadProgressCompleteFile), 1)
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(dlcManager.DLC_ON_DOWNLOAD_PROGRESS_UPDATE        ,gameManager.onDlcDownloadProgressUpdate), 1)

    -- GPU Rating Event
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(EVENT_GPU_RATING_CALCULATED                       ,gameManager.onGPURatingCalculated), 1)


    gameManager.beforeSaveFunc = beforeSaveFunc;
    gameManager.getGameDescriptionFunc = getGameDescriptionFunc or gameManager.getGameDescriptionFallbackFunc;
    gameManager.onResetFunc = onResetFunc

    persistence.init(serverManager, gameManager.getPersistenceData, PERSISTENCE_TIME, PERSISTENCE_VALIDATE, gameManager.onPersistenceSaved)
    local data = persistence.load()  --  load any existing data from the disk
    gameManager.setPersistenceData(data)
    if not table.isEmpty(data) then
        persistence.setDataSyncedWithServer()   --  mark the local data as up to date (basically sets the local data blob for diffPatcher comparisons)
    end
    
    mailboxManager.init(serverManager, MAIL_UPDATE_FREQ)

    metadataManager.init(runaway.Native:getMetaData())
    setupGameManagerConditions()
    socialManager.init(serverManager)
    publicCounterManager.init(serverManager)
    activityFeedManager.init(serverManager)

    audioManager.init(gameManager.platformData.localAudioData)
end    




function gameManager.run()
    if not DISABLE_GPU_RATING then
        gameManager.calculateGPURating()
    end

    gameManager.reset()
end

function gameManager.forceCalculateGPURating()
    gameManager.platformData.gpuScore = nil
    gameManager.platformData.gpuRating = nil
    gameManager.savePlatformSettings()
end

function gameManager.calculateGPURating()
    echoInfo("calculateGPURating")

    -- GPU Test    
    if not (gameManager.platformData.gpuRating) then

        local filePath = utils.getSaveDir(true)
        filePath = filePath .. gameManager.gpuTestDummyFilename

        -- Check if GPUTest file exists
        if cc.FileUtils:getInstance():isFileExist(filePath) == false then
            -- Create a local file called GPUTest
            local file = io.open(filePath, 'w')

            if file then
                file:close()
            end

            echoLog("GPUTest", "Calculating GPU rating")
            runaway.Native:beginCalculateGPURating()
        else
            -- File exists.
            gameManager.platformData.gpuRating = PERFORMANCE_SETTING
            gameManager.platformData.gpuScore = ERROR_GPUTEST_FAILED -- Crashed during GPU Test
            gameManager.savePlatformSettings()
            -- Send the result analytics
            gameManager.shouldFireGPUAnalytic = true

            os.remove(filePath)
        end
    end
end

function gameManager.onGPURatingCalculated(event)
    echoInfo("onGPURatingCalculated")
    local eventDataString = event:getUserDataString()
    local eventData = json.decode(eventDataString)

    gameManager.platformData.gpuScore = eventData.score
    gameManager.platformData.gpuRating = eventData.rating
    gameManager.savePlatformSettings()

    gameManager.shouldFireGPUAnalytic = true

    -- Set the effects quality to the rating
    PERFORMANCE_SETTING = gameManager.platformData.gpuRating

    -- Delete the GPUTest file if it exists
    local filePath = utils.getSaveDir(true)
    filePath = filePath .. gameManager.gpuTestDummyFilename
    if cc.FileUtils:getInstance():isFileExist(filePath) then
        os.remove(filePath)
    end
end

--  used to fully reset the game eg if an error occured
function gameManager.reset()
    echoInfo("*******************************")
    echoInfo("** full game reset triggered **")
    echoInfo("*******************************")
    gameManager.SetLoginProgress(PROGRESS_INIT)
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.StartingGame, nil);

    -- if we have new DLC then this may replace what's in the cache
    cc.FileUtils:getInstance():purgeCachedEntries();

    gameManager.authManager:setStarted(false)
    gameManager.switchScene(gameManager.SCENE_LOADING);
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.Loading, nil)
    gameManager.isStartRequested = false;
    gameManager.firstRun = true
    gameManager.serverVersion = 0
    gameManager.dlcResetRequired = false
    serverManager.setAccessToken(nil)

    socialManager.cleanup()
    audioManager.cleanup()
    errorManager.cleanup()
    dialogManager.cleanup()

    if gameManager.onResetFunc then gameManager.onResetFunc() end

    local data = persistence.load()  --  load any existing data from the disk
    gameManager.setPersistenceData(data)
    if not table.isEmpty(data) then
        persistence.setDataSyncedWithServer()   --  mark the local data as up to date (basically sets the local data blob for diffPatcher comparisons)
    end
    persistence.cleanup()
    setInitialConditionStates()

    if LOCAL_MODE and not GET_DLC_IN_LOCAL_MODE then
        --  no need to confirm with server
        conditionWatcher:setCondition(CONDITION_USER_DATA_VERIFIED, true)  
        gameManager.loadGame()
    else
        gameManager.downloadSettings()
    end
end


function gameManager.downloadSettings()
    local function onSettingsReceived(result, values)
        if result["code"] == runaway.PluginResult.kSucceed and values ~= nil then
            gameManager.SetLoginProgress(PROGRESS_SETTINGS)
            gameManager.serverSettings = values

            -- dump(values, "downloaded Settings")
            if (values.time) then
                timeManager.setServerTime(values.time)
            end

            local thisBuild = runaway.Native:getVersion()
            local dlcVersions = {}
            gameManager.serverVersion = values.version or 0
            echoInfo("Server version is currently %s", gameManager.serverVersion)

            -- currently this is only being used for splashcam snap of the day
            if values.settings and values.settings.gameSpecificSettings then
                gameManager.gameSpecificSettings = values.settings.gameSpecificSettings
            end

            -- grab the relevant values out of the settings blob
            if device.client_platform == "amazon" then
                gameManager.minVersion = values.min_version_amazon or 0
                gameManager.recommendedVersion = values.recommended_version_amazon or 0
                gameManager.maxAnalyticsVersion = values.max_analytics_version_amazon or 0
                gameManager.updateUrl = values.store_url_amazon 
                if values.settings ~= nil then
                    dlcVersions = values.settings.dlcVersionsAmazon or {}
                end
            elseif device.client_platform == "android" then
                gameManager.minVersion = values.min_version_android or 0
                gameManager.recommendedVersion = values.recommended_version_android or 0
                gameManager.maxAnalyticsVersion = values.max_analytics_version_android or 0
                gameManager.updateUrl = values.store_url_android 
                if values.settings ~= nil then
                    dlcVersions = values.settings.dlcVersionsAndroid or {}
                end
            else -- must be ios
                gameManager.minVersion = values.min_version_ios or 0
                gameManager.recommendedVersion = values.recommended_version_ios or 0
                gameManager.maxAnalyticsVersion = values.max_analytics_version_ios or 0
                gameManager.updateUrl = values.store_url_ios 
                if values.settings ~= nil then
                    dlcVersions = values.settings.dlcVersionsiOS or {}
                end
            end
            local updateDescription = (values.update_desc ~= "-") and values.update_desc or SKELETON_MESSAGES.UpdateAvailable.Desc

            echoInfo("checking version %d <==> Min:%d / Max:%d", thisBuild, gameManager.minVersion, gameManager.maxAnalyticsVersion)
            -- handle version checks here:
            if (gameManager.minVersion > thisBuild) and (gameManager.updateUrl ~= nil) then
                echoInfo("UpdateDescription = %s", (updateDescription or "MISSING"))
                dialogManager.showDialog( 
                    SKELETON_MESSAGES.UpdateAvailable.Title, 
                    updateDescription,
                    false,
                    { ui.SYSTEM_DIALOG_BUTTON_OK }, 
                    function()
                         --  send the user to the update url
                         device.openURL(gameManager.updateUrl)
                    end,
                    {hsUrl = SKELETON_MESSAGES.UpdateAvailable.hsUrl});
                return  --  don't continue with loading

             -- CHECK_RECOMMENDED_VERSION is a global constant; defined in config.lua
            elseif CHECK_RECOMMENDED_VERSION and (gameManager.recommendedVersion > thisBuild) and (gameManager.updateUrl ~= nil) then
                updateDescription = SKELETON_MESSAGES.UpdateRecommended.Desc
                echoInfo("UpdateDescription = %s", (updateDescription or “MISSING”))
                dialogManager.showDialog(
                    SKELETON_MESSAGES.UpdateRecommended.Title,
                    updateDescription,
                    false,
                    { ui.SYSTEM_DIALOG_BUTTON_CANCEL,
                      ui.SYSTEM_DIALOG_BUTTON_OK },
                        function(result)
                            if (result == ui.SYSTEM_DIALOG_BUTTON_OK) then
                                device.openURL(gameManager.updateUrl) --  send the user to the update url
                            end
                            CHECK_RECOMMENDED_VERSION = false
                            gameManager.reset()
                        end,
                    {hsUrl = SKELETON_MESSAGES.UpdateRecommended.hsUrl});
                return --  don’t continue with loading
            end

            --  find the requried DLC version
            local requiredDlcVersion = 0
            --  sort the dlcVersions table in ascending order of "maxBuild" property
            table.sort(dlcVersions, function(lhs,rhs) return (lhs.maxBuild or 0) < (rhs.maxBuild or 0) end)
            for i = 1,#dlcVersions do
                local version = dlcVersions[i]
                if version.maxBuild <= thisBuild then
                    requiredDlcVersion = version.dlcVersion
                end
            end
            echoInfo("DLC Version required for Version %d = %s", thisBuild, requiredDlcVersion)

            if values.dlc_url ~= nil then
                gameManager.dlc_url = values.dlc_url
                gameManager.posts_res_url = gameManager.dlc_url.."posts/"
            end

            -- check if we have a DLC override
            if device.isLauncherAppInstalled() then
                gameManager.isDeveloperDevice = true    --  use this flag to see if we are a developer device.
                local metaData = metadataManager.getMetadata()
                if metaData and metaData.dlcVersion then
                    requiredDlcVersion = metaData.dlcVersion
                    gameManager.maxAnalyticsVersion = 0             -- disable analytics  
                    gameManager.displayDlcOverrideVersion(requiredDlcVersion, true)
                else
                    gameManager.displayDlcOverrideVersion(requiredDlcVersion, false)
                end
            end


            if gameManager.dlc_url ~= nil and requiredDlcVersion ~= nil then
                dlcManager:setServer(values.dlc_url, requiredDlcVersion)
                local fullDlcRequired = gameManager.authDataExists()
                dlcManager:startSync(nil, fullDlcRequired or DLC_ALWAYS_BLOCKING)
                --  this will eventually call gameManager.onDlcDownloadCompleteSuccess(), 
                --  which will call gameManager.loadGame()
            else
                -- no DLC
                gameManager.loadGame()
            end

        else
            dialogManager.showDialog( 
                SKELETON_MESSAGES.NetworkFailedSettings.Title, 
                SKELETON_MESSAGES.NetworkFailedSettings.Desc,
                true,
                { ui.SYSTEM_DIALOG_BUTTON_OK }, 
                gameManager.reset,
                {hsUrl = SKELETON_MESSAGES.NetworkFailedSettings.hsUrl});
        end
    end

    echoInfo("Downloading server settings")
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.DownloadingSettings, nil);
    serverManager.performRequest("Settings", nil, nil, onSettingsReceived, true)
end

function gameManager.authDataExists()
    local authData = gameManager.authState.load()   --  check if we have previously saved auth data
    return (LOCAL_MODE == true) or (authData ~= nil)
end

function gameManager.getAuthData()
    return gameManager.authState.load()
end

function gameManager.saveUserAuthData(socialType, userID)
    -- generate the authentication file name
    local writablePath = cc.FileUtils:getInstance():getWritablePath()
    local filename = string.gsub(writablePath, "[\\\\/]+$", "") .. "/" .. "authentication"
    local authData = {}
    authData.userId = userID
    authData.accountType = socialType
    ret = io.writefile(filename, json.encode(authData))
end

function gameManager.loadGame()
    --  delay the initial run because the event listeners don't work until the next update.
    scheduler.performWithDelayGlobal(
        function()
            if gameManager.authDataExists() then
                echoInfo("Account information exists, just resume the game")
                if (LOCAL_MODE) then
                    gameManager.enterGameScene()
                else
                    gameManager.SetUpdateMessage(SKELETON_MESSAGES.LoggingIn, nil);

                    --determine the social type to log in with.
                    authData = gameManager.authState.load()
                    if authData and authData.accountType then
                        echoInfo("Attempting to log in user with accountType=%d", authData.accountType)
                        gameManager.loginSocial(authData.accountType)
                    else
                        echoInfo("unsure of method to log in user.  cleaning out local data and resetting")
                        gameManager._deleteLocalData()
                    end
                end
            else
                gameManager.SetUpdateMessage(SKELETON_MESSAGES.NoSession, nil)
                gameManager.switchScene(gameManager.SCENE_START);
            end
        end, 0)
end


--  used to trigger an ingame restart (eg if DLC is updated, or if server data is out of sync)
--  this restart is not as heavy handed as reset
function gameManager.restartGame()
    echoInfo("Restart triggered")
    conditionWatcher:setCondition(CONDITION_DATA_UNSAFE,        false)
    gameManager.authManager:resume()
    if gameManager.isOnEnterForeground and gameManager.getIsInGame() then
        --  ignore the restart, because we are already in game. (this is for the minimise/restore scenario)
        echoInfo("Skipping Restart upon entering foreground")
    else
        -- if we have new DLC then this may replace what's in the cache
        cc.FileUtils:getInstance():purgeCachedEntries();
    
        gameManager.switchScene(SCENE_RESET);--  to ensure that we do actually reset if we're currently in game
        if gameManager.onResetFunc then gameManager.onResetFunc() end
        gameManager.switchScene(gameManager.SCENE_GAME); 
    end
    gameManager.isOnEnterForeground = false;
end

function gameManager.linkSocialAccount()
    local function onAccountLinked(result, data)
        if result.code == runawayplay.PluginResult.kSucceed then
            --  TODO: deal with newly added account
        end
    end
    gameManager.authManager:linkAccount(onAccountLinked)
end


function gameManager.startNewGuestGame()
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.StartNewGame, nil)
    gameManager.switchScene(SCENE_RESET)
    gameManager.isNewAccount     = true;    --  flag so that we save the data upon login.
    gameManager.startLogin(nil, SOCIAL_GUEST)
end


function gameManager.loginSocial(socialType)

    local function onSocialLoggedIn(result, data)
        -- dump(result, "onSocialLoggedIn result")
        -- dump(data, "onSocialLoggedIn data") --  data will be the result of the profile request

        if result.code == runaway.PluginResult.kSucceed then
            data = data or {}
            if data.socialAccessToken ==  "" or data.socialAccessToken == nil then
                data.socialAccessToken = socialManager.getAccessToken(socialType)
            end
            if data.socialAccessToken then
                gameManager.startLogin(data.socialAccessToken, socialType)
            else
                gameManager.onLoginFailed()
            end
        else
            gameManager.onLoginFailed()
        end
    end

    socialType = tonumber(socialType) -- fix for android treating things as strings

    if      socialType == SOCIAL_GUEST 
        or  socialType == SOCIAL_LCD 
        or  socialType == SOCIAL_MOBAGE
        or  socialType == SOCIAL_GAMECENTER
        or  socialType == SOCIAL_GOOGLEPLAY 
        or  socialType == SOCIAL_AMAZON then
        
        gameManager.startLogin(nil, socialType, callback)
    else
        local socialService = socialManager.getSocialServiceDescription(socialType)
        socialManager.loginService(socialService, onSocialLoggedIn)
    end
end


--  called by gameManager.authManager when the LCD session is updated/created
--  this will connect to the Game Server and get an access token for us to talk to it with.
--      Note, this should only get called if using LCD, 
--      or the first time that we log in after turning off LCD
function gameManager.onNewLcdSession(sessionData)
    local retry = nil
    local attempts = 0
    echoInfo("gameManager.onNewLcdSession() (called from gameManager.authManager.LcdOnSessionUpdate)")

    --  used to handle LCD LoadAcccount on android
    if type(gameManager.onLcdResumed) == "function" then
        echoInfo("Running delayed function upon LCD resume")
        local funcToRun = gameManager.onLcdResumed
        gameManager.onLcdResumed = nil
        funcToRun()
        return
    end

    if gameManager.authManager:getIsPaused() then
        echoInfo("gameManager.authManager is paused, so ignoring")
        return
    end

    lcdUserId = sessionData and sessionData.user and sessionData.user.userId or -1
    -- dump(authData, "authData in gameManager.onNewLcdSession()")
    echoInfo("comparing userIds LCD:%s <==> Current:%s", lcdUserId, gameManager.authUserId)
    echoInfo("  serverManager.hasAccessToken:%s", serverManager.hasAccessToken() and "Y" or "N")
    if (lcdUserId == -1) then
        dump(sessionData, "unexpected sessionData in onNewLcdSession - terminating login")
        return
    end    

    -- only perform if:
    --  we're not in game
    --  or our user has changed
    --  or we have no oauth token (ie: aren't yet logged in)
    local isInGameScene = gameManager.isSceneInGame()
    if not isInGameScene or (lcdUserId ~= gameManager.authUserId) or (not serverManager.hasAccessToken()) then
        echoInfo("running LoginLcd because:")
        if not isInGameScene then
            echoInfo(" - we're NOT in game")
        end
        if (lcdUserId ~= gameUserId) then
            echoInfo(" - the userId HAS changed")
        end
        if (lcdUserId ~= gameUserId) then
            echoInfo(" - we don't have an access token")
        end

        gameManager.lcdSessionData = sessionData
        gameManager.startLogin(sessionData.accessToken, SOCIAL_LCD)
    else
        echoInfo("skipping LoginLcd because:\n - user hasn't changed\n - we're in game\n - we have an access token")
    end
end

local function makeGuid()
    local random = math.random
    local template ='xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'
    return string.gsub(template, '[xy]', function (c)
        local v = (c == 'x') and random(0, 0xf) or random(8, 0xb)
        return string.format('%x', v)
    end)
end            


function gameManager.startLogin(accessToken, socialType, callback)
    local function onLoginNativeComplete(result, data)
        -- dump(result, "onLoginNativeComplete result")
        -- dump(data, "onLoginNativeComplete data")
        if type(callback) == "function" then
            callback(result, data)
        end
        gameManager.onLoginComplete(result, data)
    end

    local params = {}
    params["access_token"] = accessToken
    local userId = nil
    local storeUserId = nil
    local authData = gameManager.authState.load()
    if authData and authData.userId then
        userId = authData.userId
        storeUserId = authData.storeAccount and authData.storeAccount.storeUserId or userId
    else
        userId = makeGuid()         -- probably should do a check against the server to see if this accountId already exists
        storeUserId = runaway.Native:getDeviceIdentifier()
    end
    params["userId"] = userId
    params["storeUserId"] = storeUserId
    params["socialType"] = socialType
    dump(params, "gameManager.startLogin (Native)->params")

    serverManager.performRequest(gameManager.authManager.LoginEndpoint, params, nil, onLoginNativeComplete, true)
end

function gameManager.startLoginWithAccount(userId, socialType, callback)
    local function onLoginWithAccountComplete(result, data)
        dump(result, "onLoginWithAccountComplete result")
        dump(data, "onLoginWithAccountComplete data")
        if type(callback) == "function" then
            callback(result, data)
        end
        gameManager.onLoginComplete(result, data)
    end

    --   if we are trying to login with a login-able social account, then use the correct mechanism
    if socialType == SOCIAL_FACEBOOK then --or socialType == SOCIAL_GOOGLEPLAY or socialType == SOCIAL_GAMECENTER then
        gameManager.loginSocial(socialType, callback)
    else
        local params = {}
        params["userId"] = userId
        params["socialType"] = socialType
        dump(params, "gameManager.startLogin (Native)->params")
        serverManager.performRequest(gameManager.authManager.LoginEndpoint, params, nil, onLoginWithAccountComplete, true)
    end
end

function gameManager.onLoginFailed()
    dialogManager.showDialog( 
        SKELETON_MESSAGES.NetworkFailedLogin.Title, 
        SKELETON_MESSAGES.NetworkFailedLogin.Desc,
        true,
        { ui.SYSTEM_DIALOG_BUTTON_OK }, 
        gameManager.reset,
        {hsUrl = SKELETON_MESSAGES.NetworkFailedLogin.hsUrl});    
end

function gameManager.onLoginComplete(result, data)
    dump(result, "onLoginComplete.result")
    dump(data, "onLoginComplete.data")
    dump(gameManager.lcdSessionData, "onLoginComplete.lcdSessionData")

    local migrationError = serverManager.hasSpecificError(data, 34)

    if migrationError then
        gameManager.authState.delete()  
        dialogManager.showDialog( 
            SKELETON_MESSAGES.MigrationFailed.Title, 
            SKELETON_MESSAGES.MigrationFailed.Desc,
            true,
            { ui.SYSTEM_DIALOG_BUTTON_OK }, 
            gameManager.reset,
            {SKELETON_MESSAGES.MigrationFailed.hsUrl});
        return
    end

    if (result.code == runaway.PluginResult.kSucceed) then
        gameManager.SetLoginProgress(PROGRESS_LOGGED_IN)

        echoInfo("onLoginComplete succeeded")
        --  data.access_token is our new OAuth access token
        --  gameManager.lcdSessionData.user is our LCD user data
        serverManager.setAccessToken(data['access_token']) 
        local userData = {}

        userData = data.userAccount or {}
        userData.userId = userData.userId or userData.accountId
        gameManager.authState.save(userData)

        gameManager.authUserId = userData.userId
        gameManager.getUserFromServer()
    else
        echoInfo("onLoginComplete failed!")
        dump(gameManager.lcdSessionData, "lcdSessionData")
        dump(persistenceUser, "persistenceUser")
        gameManager.onLoginFailed()
    end
--        gameManager.lcdSessionData.userId
end

function gameManager.logout()
    local function onAccountsLoggedOut()
        gameManager.authState.delete()  
        gameManager.lcdSessionData = nil    
        gameManager.authManager.authData = nil
        persistence.delete()
        platformUserGameData.clearData()
        gameManager.reset()
    end
    
    local function onForceSaved(result, data)
        socialManager.logoutAllAccounts(onAccountsLoggedOut)
    end

    local function onLogoutConfirmed(button)
        if (button == ui.SYSTEM_DIALOG_BUTTON_YES) then
            persistence.force(nil, nil, true, onForceSaved)
        end
    end 

    --  user this mechanism in the future when we can log in with GameCenter/GooglePlay
    -- local serviceDescription = socialManager.getMainSocialService()
    -- if serviceDescription then

    --  until that point, just see if we are logged in to facebook
    if socialManager.isSocialServiceActive(SOCIAL_FACEBOOK) then
        -- determine which type of social network we've logged in with 
        local serviceDescription = socialManager.getSocialServiceDescription(SOCIAL_FACEBOOK)
        dialogManager.showDialog(
                                SKELETON_MESSAGES.LogoutSocial.Title,
                                string.format(SKELETON_MESSAGES.LogoutSocial.Desc, serviceDescription.name),
                                true,
                                { ui.SYSTEM_DIALOG_BUTTON_CANCEL, ui.SYSTEM_DIALOG_BUTTON_YES },
                                onLogoutConfirmed,
                                {hsUrl = SKELETON_MESSAGES.LogoutSocial.hsUrl}
                            )
    else
        dialogManager.showDialog(
                                SKELETON_MESSAGES.LogoutGuest.Title,
                                SKELETON_MESSAGES.LogoutGuest.Desc,
                                true,
                                { ui.SYSTEM_DIALOG_BUTTON_CANCEL, ui.SYSTEM_DIALOG_BUTTON_YES },
                                onLogoutConfirmed,
                                {hsUrl = SKELETON_MESSAGES.LogoutGuest.hsUrl}
                            )
    end
end


function gameManager.getUserFromServer()
    local retry = nil
    local attempts = 0

    local function onGetUser(result, data)

        -- TODO: remove this when the admin portal is working
        
        local data = ONGETUSER_DATA(data)

        if result.code == runaway.PluginResult.kSucceed then
            gameManager.SetLoginProgress(PROGRESS_USERDATA)
            if (data.time) then
                echoInfo("Updating time from server (ONGETUSER)")
                timeManager.setServerTime(data.time)
            end
            if not gameManager.isSceneInGame() then
                echoInfo("Loading game data from server on app startup")
                gameManager.setPersistenceData(data)
                gameManager.gameDataLoadCompleted(data)
            else
                echoInfo("Loading game data from server after link/load")
                gameManager.onLocalDataOutOfDate(data)
            end
        else
            dialogManager.showDialog( 
                SKELETON_MESSAGES.NetworkFailedGetUser.Title, 
                SKELETON_MESSAGES.NetworkFailedGetUser.Desc,
                true,
                { ui.SYSTEM_DIALOG_BUTTON_OK }, 
                gameManager.reset,
                {hsUrl = SKELETON_MESSAGES.NetworkFailedGetUser.hsUrl});
        end
    end
   
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.DownloadingGameData, nil);
    serverManager.performRequest("User", nil, nil, onGetUser, true)
end


function gameManager.onLocalDataOutOfDate(data)
    echoInfo("gameManager.onLocalDataOutOfDate in scene ".. gameManager.currentScene)
    --conditionWatcher:setCondition(CONDITION_USER_DATA_VERIFIED, false)   

    gameManager.isOnEnterForeground = false;
    if gameManager.isSceneInGame() then
        echoInfo("Updating obsolete local data - in game")
        conditionWatcher:setCondition(CONDITION_DATA_UNSAFE, true)   
        gameManager.setPersistenceData(data)
        persistence.setDataSyncedWithServer()
        gameManager.reset()
    else
        echoInfo("Updating obsolete local data - out of game")
        gameManager.setPersistenceData(data)
        persistence.setDataSyncedWithServer()
        gameManager.gameDataLoadCompleted(data)
    end
end

function gameManager.checkGameDataVersion(data)
    -- check that the GAME_VERSION is not lower than the version registered in the data.
    dump(persistenceUserGame, "gameManager.checkGameDataVersion")

    local saveVersion = persistenceUserGame.public.saveVersion or 0
    echoInfo("saveVersion = %s, gameVersion = %s", saveVersion, GAME_VERSION)
    if (saveVersion > GAME_VERSION) then
        conditionWatcher:setCondition(CONDITION_DATA_UNSAFE, true)
        dialogManager.showDialog( 
            SKELETON_MESSAGES.UpdateRequired.Title, 
            SKELETON_MESSAGES.UpdateRequired.Desc,
            true,
            { ui.SYSTEM_DIALOG_BUTTON_OK }, 
            function()
                --  send the user to the update url
                if (gameManager.updateUrl) then
                    device.openURL(gameManager.updateUrl)
                else
                    gameManager.reset()
                end
            end,
            {hsUrl = SKELETON_MESSAGES.UpdateRequired.hsUrl});
        return false --  don't continue with loading
    end

    --  stamp this version.
    persistenceUserGame.public.saveVersion = GAME_VERSION
    return true;
end

function gameManager.gameDataLoadCompleted(data)
    echoInfo("gameManager.gameDataLoadCompleted()")
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.LoadingComplete, nil);
    gameManager.SetLoginProgress(PROGRESS_GAME)
    conditionWatcher:setCondition(CONDITION_USER_DATA_VERIFIED, true)
    conditionWatcher:setCondition(CONDITION_DATA_UNSAFE, false)

    local publicData = data and data.userGame and data.userGame.public
    gameManager.userGameId = data and data.userGame and data.userGame.id
    gameManager.userId = data and data.user and data.user.id
    echoInfo("Loaded userGameId %s", gameManager.userGameId or "MISSING")
    if gameManager.userGameId and gameManager.userId then 
        if (type(gameManager.onUserLoggedInFunc) == "function") then
            gameManager.onUserLoggedInFunc(gameManager.userGameId, gameManager.userId)
        end 
        gameManager.savePlatformSettings()        
    end

    local gameScore = 0
    if publicData ~= nil then
        gameScore = gameManager.getScoreFunc(publicData)
    end

    if gameManager.checkGameDataVersion(data) then
        socialManager.updateUser(data)

        platformUserGameData.checkSponsorId()

        if (data and data.mailItems) then
            -- record any mailbox items sent through with the user
            mailboxManager.handleMailFromServer(data.mailItems)
        end

        if (gameScore > 0) and (dlcManager:isDownloadInProgress()) then
            echoInfo("Game data loaded, but non-blocking DLC is still loading.  skipping game start with gameScore %s", gameScore)
            gameManager.dlcResetRequired = true
            -- switch to the loading scene to continue with DLC loading
            gameManager.switchScene(gameManager.SCENE_LOADING)
            gameManager.SetLoginProgress(PROGRESS_DLC)
            gameManager.SetUpdateMessage(SKELETON_MESSAGES.DownloadingDlc, "");
        else
            gameManager.enterGameScene()            --  alternatively, this is called from onDlcDownloadCompleteSuccess
        end
    end
end



function gameManager.enterGameScene()
    echoInfo("gameManager.enterGameScene()")


    if not gameManager.isSceneInGame() then
        gameManager.switchScene(gameManager.SCENE_GAME)
        --  also ensure that GameCenter/GooglePlay is loaded

        if not LOCAL_MODE then
            scheduler.performWithDelayGlobal(
                function()
                    socialManager.updateSocial()
                    --  no longer ensures that GameCenter/GooglePlay is loaded here.
                    --  now the game code calls this after tutorial is completed
                    -- gameManager.autoLoadBasicSocial()
                end, 5) --  wait 5 seconds
        end
    else
        gameManager.restartGame();
    end
end

function gameManager.autoLoadBasicSocial()
    if not gameManager.getPlatformSettings().hasAutoLoadedBasicSocial then
        socialManager.autoLoadBasicSocial()
        gameManager.hasAutoLoadedBasicSocial = true
        gameManager.savePlatformSettings()
    end
end

function gameManager.linkAccount()
    echoInfo("gameManager.linkAccount()")
    local function onLinkAccount(result, data)
        dump(result, "onLinkAccount[result]")
        dump(data, "onLinkAccount[data]")
        if result.code == runaway.PluginResult.kSucceed then
            persistenceUserGame.public.isLinked = true
        end
        -- no intelligent information added which will tell us which social network was linked 
    end

    gameManager.authManager:linkAccount(onLinkAccount)
end

function gameManager.loadAccount(callback)
    if not gameManager.getIsInGame() then
        local loginMenu = require('layers.loginMenuLayer').build(callback)
        gameManager.currentSceneInstance:addChild(loginMenu, 10)   
    end
end


function gameManager.handleLostSession()
    -- persistence.stop()
    conditionWatcher:setCondition(CONDITION_USER_DATA_VERIFIED, false)   -- this will also stop persistence
    serverManager.setAccessToken(nil)
    gameManager.reset() --  now performs a full reset on lost session
    -- gameManager.authManager:resume()
end


--  checks the Game Server session (will be called upon minimise/restore)
function gameManager.checkSessionLocal()
    -- need to re-implement
    local function onLocalSessionCheckComplete(result, data)
        --determine if there was any errors, (eg: an error #1 is stating that the sesison was re-claimed)
        local hasErrors = false
        if (data ~= nil) and (type(data.errors) == "table") and not table.isEmpty(data.errors) then
            hasErrors = true
        end

        if (result.code == runaway.PluginResult.kSucceed) and not hasErrors then
            --   all good in the hood
            echoInfo("Local Session checks out valid")
            if (data.time) then
                timeManager.setServerTime(data.time)
            end
            conditionWatcher:setCondition(CONDITION_USER_DATA_VERIFIED, true)
            cc.Director:getInstance():getEventDispatcher():dispatchEvent(cc.EventCustom:new(EVENT_ON_SESSION_CHECK))
        else
            echoInfo("Local Session is not valid")
            dump(result, "onLocalSessionCheckComplete->result")
            dump(data, "onLocalSessionCheckComplete->data")
            gameManager.handleLostSession()
        end
    end

    if LOCAL_MODE then
        return
    else
        conditionWatcher:setCondition(CONDITION_USER_DATA_VERIFIED, false)
        echoInfo("Local Session check")
        serverManager.performRequest("Session", {}, nil, onLocalSessionCheckComplete)
    end 
end


function gameManager.getGameDescriptionFallbackFunc(data)
    --fallback function if one isn't specified
    local gameDesc = "Unknown"
    if data and data.userGame and data.userGame.public then
        local score = data.userGame.public.score
        local created = os.date("%c", data.userGame.created_on)
        local lastPlayed = os.date("%c", data.userGame.updated_on)
        gameDesc = string.format("Score: %d\nCreated: %s\nLast Played: %s", score or 0, created, lastPlayed)
    end
    echoInfo("GameDescription:\n%s", gameDesc)
    return gameDesc
end



function gameManager.recoverFromError() -- Called when a Lua error is thrown
    echoInfo("RECOVERING FROM ERROR")
    gameManager.reset()
end











--  ************************************************
--  *     Persistence handling section starts      *
--  ************************************************

function gameManager.getPersistenceData(ignoreBeforeSaveFunc)
    --  update user data if required here

    if (not ignoreBeforeSaveFunc) and (gameManager.beforeSaveFunc ~= nil) then
        gameManager.beforeSaveFunc()
    end

    persistenceData.platformData = platformUserGameData.getData()

    return persistenceInternal
end
function gameManager.setPersistenceData(data)
    if (type(data) ~= "table") then
        echoError("ERROR Setting persistenceData to nonTable type:".. type(data))
    end

    if not gameManager.firstRun and table.isEmpty(data) then
        echoError("ERROR Setting persistenceData to Empty Data list")
    end

    --dump(data, "persistence saved data")
    persistenceInternal =   clone(data or {});
    persistenceInternal.errors = nil;-- ignore errors node from downloaded data
    persistenceInternal.reset = nil;-- to specifically allow reset data
    --  update access variables

    --  ensure that the required nodes exist
    persistenceInternal.user = persistenceInternal.user or {}
    persistenceInternal.userGame = persistenceInternal.userGame or {}
    persistenceInternal.userGameBucket = persistenceInternal.userGameBucket or {}

    persistenceUser =       persistenceInternal.user
    persistenceUserGame =   persistenceInternal.userGame
    persistenceData =       persistenceInternal.userGameBucket

    platformUserGameData.restoreData(persistenceData.platformData)

    --  ensure that editable user tables are created correctly
    persistenceUser.data = persistenceUser.data or {}
    persistenceUserGame.public = persistenceUserGame.public or {}
    persistenceUserGame.public.activeAccounts = persistenceUserGame.public.activeAccounts or {}

    -- put anything you want to share between the games in here
    persistenceUser.data[APP_NAME] = persistenceUser.data[APP_NAME] or {}
    persistenceUser.data[APP_NAME].lastLoginTimestamp = timeManager.getTime()
end

function gameManager.onPersistenceSaved()
    echoInfo("Persistence has successfully saved to server")
    mailboxManager.outwardsMailQueue.flushMail()
end


function gameManager._deleteLocalData()
    saveManager.cleanupLocal()
    gameManager.setPersistenceData({reset=1})
    gameManager.reset()
end


function gameManager.deleteServerData()

    if (LOCAL_MODE) then
        dialogManager.showDialog( --layer to attach to, (nil == MENULAYER)
            "!!!WARNING!!!", 
            "This will delete ALL of your local data.  You are in local mode, so no server data will be changed\nDo you wish to continue?",
            true,
            {  ui.SYSTEM_DIALOG_BUTTON_CANCEL, ui.SYSTEM_DIALOG_BUTTON_OK }, 
            function(result)
                if (result ==  ui.SYSTEM_DIALOG_BUTTON_OK) then
                    gameManager._deleteLocalData()
                end
            end);
    else
        function onDeleted(result)
            dump(result, "deleteServerData->onDeleted")
            if (result.code == runaway.PluginResult.kSucceed) then
                gameManager._deleteLocalData();
            else
                dialogManager.showDialog( --layer to attach to, (nil == MENULAYER)
                    "!!!WARNING!!!", 
                    "Something failed during delete.  Your data is likely in an inconsistent state, so you should hard reset",
                    true,
                    {  ui.SYSTEM_DIALOG_BUTTON_OK}, 
                    function() end);
            end
        end

        dialogManager.showDialog( --layer to attach to, (nil == MENULAYER)
            "!!!WARNING!!!", 
            "This will delete ALL of your user data on the server, as well as your local client\nDo you wish to continue?",
            true,
            {  ui.SYSTEM_DIALOG_BUTTON_CANCEL, ui.SYSTEM_DIALOG_BUTTON_OK }, 
            function(result)
                if (result ==  ui.SYSTEM_DIALOG_BUTTON_OK) then
                    conditionWatcher:setCondition(CONDITION_DATA_UNSAFE,     true)   --  will pause persistence etc 
                    serverManager.performRequest("UserDelete", params, nil, onDeleted)
                end
            end);
    end
end

--  ************************************************
--  *       Scene loading section starts           *
--  ************************************************

function gameManager.switchScene(sceneName)
    if (sceneName) then
        if (gameManager.currentScene ~= sceneName) then
            echoInfo("switchScene From ".. (gameManager.currentScene or "").." To ".. sceneName)
            -- echo(debug.traceback("", 2))

            gameManager.currentScene = sceneName;
            local scene = require(SCENE_PATH.."."..sceneName).create(sceneName)
            gameManager.currentSceneInstance = scene
            scene:showWithScene()
        end
    else
        gameManager.currentScene = ""
        gameManager.currentSceneInstance = nil
        echoInfo("Attempt to switch to null scene")
        -- echoError("Attempt to switch to null scene")
    end
end

function gameManager.isSceneInGame()
    local inGame = (gameManager.currentScene == gameManager.SCENE_GAME)
    echoInfo(inGame and "Yes, we're in game" or "No, we're not in game")
    return inGame;
end

function gameManager.isSceneInStart()
    local inStart = (gameManager.currentScene == gameManager.SCENE_START) or (gameManager.currentScene == gameManager.SCENE_LOADING)
    echoInfo(inStart and "Yes, in start scene" or "No, we are not in start scene")
    return inStart
end


function gameManager.setRunOnInGameScene(func)
    --  if we're in game, then just run the function
    if (gameManager.getIsInGame()) then
        func();
    else
        --  else, we're likely in the game loading scene, so store the function to run once the game loads.
        gameManager.runOnInGameScene = func;
    end
end

function gameManager.setIsInGame(isInGame)
    gameManager._isInGame = isInGame
    conditionWatcher:setCondition(CONDITION_GAME_IS_RUNNING,    isInGame)

    if (isInGame) then
        if (type(gameManager.runOnInGameScene) == "function") then
            scheduler.performWithDelayGlobal(function()
                gameManager.runOnInGameScene();
                gameManager.runOnInGameScene = nil;
            end, 0)
        end
    end
end

function gameManager.getIsInGame()
    return gameManager._isInGame or false;
end

function gameManager.onEnterBackground()
    timeManager.pauseTime()
    gameManager.savePlatformSettings()
end

function gameManager.savePlatformSettings()
    --coalate and save all platform related settings here
    gameManager.platformData.localAudioData = audioManager.getLocalSaveData()
    if gameManager.hasAutoLoadedBasicSocial then
        gameManager.platformData.hasAutoLoadedBasicSocial = true
    end
    if gameManager.authUserId then
        gameManager.platformData.authUserId = gameManager.authUserId
    end
    if gameManager.userGameId then
        gameManager.platformData.userGameId = gameManager.userGameId
    end
    dump(gameManager.platformData, "saving gameManager.platformData")
    gameManager.platformState.save(gameManager.platformData)
end

function gameManager.getPlatformSettings()
    return gameManager.platformData or {}
end

function gameManager.displayGameIdOnLayer(layer, font, position, size, labelMaker)
    -- see if we have a userGameId and display it on the screen.
    local userGameId = gameManager.userGameId or gameManager.getPlatformSettings().userGameId 
    if userGameId then
        echoInfo("displaying GameId: %s", userGameId)
        local labelText = string.format("ID: %s", userGameId)
        local size = size or 10
        echoInfo(labelText)
        local userGameIdLabel 
        if position then
            if labelMaker then
                local labelArgs = {
                    string = labelText, 
                    font = font, 
                    fontSize = size, 
                    hAlign = cc.TEXT_ALIGNMENT_CENTER, 
                    vAlign = cc.VERTICAL_TEXT_ALIGNMENT_CENTER
                }
                userGameIdLabel = labelMaker(labelArgs)
            else
                userGameIdLabel = cc.Label:createWithTTF(labelText, font, size, cc.size(display.width,0), cc.TEXT_ALIGNMENT_CENTER, cc.VERTICAL_TEXT_ALIGNMENT_CENTER)
            end
            userGameIdLabel:setAnchorPoint(cc.p(0.5,0.5))
            userGameIdLabel:setPosition(position)
        else
            -- default to top left
             if labelMaker then
                local labelArgs = {
                    string = labelText, 
                    font = font, 
                    fontSize = size, 
                    hAlign = cc.TEXT_ALIGNMENT_LEFT, 
                    vAlign = cc.VERTICAL_TEXT_ALIGNMENT_CENTER
                }
                userGameIdLabel = labelMaker(labelArgs)
            else
                userGameIdLabel = cc.Label:createWithTTF(labelText, font, size, cc.size(display.width,0), cc.TEXT_ALIGNMENT_LEFT, cc.VERTICAL_TEXT_ALIGNMENT_TOP)
            end
            userGameIdLabel:setAnchorPoint(cc.p(0.0,1.0))
            userGameIdLabel:setPosition(cc.p(2, display.height - 2))
        end
        layer:addChild(userGameIdLabel)
        return userGameIdLabel
    end
end

function gameManager.sendDebugLog(item)
    --only one debug message for now, but in the future item can have the following parameters (plus any others as required):
    --  item.amount
    --  item.data
    --  item.mailType = 2
    --  item.message
    -- dump(item, "sendDebugLog.item")
    if item.data == "apilog" then
        -- local message = item.message or "no Message"
        -- local errorDetail = {}
        -- errorDetail[1] = "Logging conditions " .. message
        -- errorDetail[2] = conditionWatcher:getConditionText()
        -- sendLogToServer(errorDetail)
    else
        local message = item.message or "no Message"
        local errorDetail = {}

        errorDetail[1] = "Logging conditions #1 " .. message
        errorDetail[2] = conditionWatcher:getConditionText()
        sendLogToServer(errorDetail)

        errorDetail[1] = "Logging AuthData #2 " .. message
        errorDetail[2] = json.encode(gameManager.getAuthData() or {})
        sendLogToServer(errorDetail)

        errorDetail[1] = "Logging traceback #3 " .. message
        errorDetail[2] = debug.traceback()
        sendLogToServer(errorDetail)
    end
end

--  ************************************************
--  *       Scene loading section ends             *
--  ************************************************



--  ************************************************
--  *        DLC Thread section starts             *
--  ************************************************

-- TODO: add better feedback messages etc



gameManager.dlcPercentComplete = 0
gameManager.loginPercentComplete = 0

function gameManager.onDlcDownloadStart(event) -- (isBlocking)
    print("download starting")
    local data = json.decode(event._usedata);
--    blockingLabel:setVisible(data.isBlocking);

    gameManager.dlcPercentComplete = 0;
    -- local majorMessage = data.isBlocking and "Checking for required updates" or "Checking for background updates"
    local majorMessage = SKELETON_MESSAGES.DownloadingDlc
    local minorMessage = SKELETON_MESSAGES.DownloadingPercentage
    local completed, total = dlcManager:getDlcCounts()
    
    -- Show download progress of blocking dlc instead if DLC_BLOCKING_PROGRESS_FIRST
    -- flag is set AND there is blocking dlc remaining.
    -- DLC_BLOCKING_PROGRESS_FIRST is a global constant; for Splash this is
    -- defined in config.lua
    if DLC_BLOCKING_PROGRESS_FIRST and dlcManager:blockingDlcRemains() then
        completed, total = dlcManager:getBlockingDlcCounts()
    end

    if completed > total then
        completed = total
    end
    if total == 0 then
        --  don't show the file count when just downloading the DLC manifest
        gameManager.SetUpdateMessage(majorMessage, "");
    else
        gameManager.SetUpdateMessage(majorMessage, minorMessage.." "..(completed + 1).."/"..total);
    end    
    gameManager.SetLoginProgress(PROGRESS_DLC, {completed = completed, total = total})
end;

function gameManager.onDlcDownloadCompleteSuccess(event) -- (isBlocking, successes, failures)
    local data = json.decode(event._usedata);
    echoInfo("gameManager.onDlcDownloadCompleteSuccess(Required=%s)", (data.isBlocking and "true" or "false"))
    gameManager.dlcPercentComplete = 100;
    
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.UpdateComplete, "");
    --  needs: AND isGameStarted


    if (not data.isBlocking) then
        if gameManager.dlcResetRequired then
            -- gameManager.reset()
            gameManager.enterGameScene()
        end
    else
        gameManager.loadGame()
    end
end;

function gameManager.onDlcDownloadCompleteError(event) --(errorMessage, isBlocking)
    local data = json.decode(event._usedata);
    --blockingLabel:setVisible(data.isBlocking);
    gameManager.SetUpdateMessage(SKELETON_MESSAGES.UpdateCompleteWithErrors, data.errorMessage);
    local function restartSync(button)
        dlcManager:startSync(nil, not gameManager.firstRun or DLC_ALWAYS_BLOCKING);
    end
    dialogManager.showDialog( 
                    SKELETON_MESSAGES.NetworkFailedDlc.Title, 
                    SKELETON_MESSAGES.NetworkFailedDlc.Desc,
                    true,
                    { ui.SYSTEM_DIALOG_BUTTON_OK }, 
                    restartSync,
                    {hsUrl = SKELETON_MESSAGES.NetworkFailedDlc.hsUrl});
end;

function gameManager.onDlcDownloadProgressCompleteFile(event) -- (file, isBlocking)
    gameManager.dlcPercentComplete = 100;
    gameManager.SetUpdateMessage(nil, nil);
end;

-- rounds a number to the nearest decimal places
function round(val, decimal)
  if (decimal) then
    return math.floor( (val * 10^decimal) + 0.5) / (10^decimal)
  else
    return math.floor(val+0.5)
  end
end

function gameManager.onDlcDownloadProgressUpdate(event) --(downloadedBytes, totalBytes, isBlocking)
    local data = json.decode(event._usedata);
    local totalBytes = (data.totalBytes ~= -1) and data.totalBytes or data.downloadedBytes
    gameManager.dlcPercentComplete = round((data.downloadedBytes * 100 / totalBytes), 0);

    local majorMessage = SKELETON_MESSAGES.DownloadingPercentage
    local completed, total = dlcManager:getDlcCounts()
    
    -- Show download progress of blocking dlc instead if DLC_BLOCKING_PROGRESS_FIRST
    -- flag is set AND there is blocking dlc remaining.
    -- DLC_BLOCKING_PROGRESS_FIRST is a global constant; for Splash this is
    -- defined in config.lua
    if DLC_BLOCKING_PROGRESS_FIRST and dlcManager:blockingDlcRemains() then
        completed, total = dlcManager:getBlockingDlcCounts()
    end

    if total == 0 then
        --  don't show the file count when just downloading the DLC manifest
        gameManager.SetUpdateMessage(nil, "")
    else
        gameManager.SetUpdateMessage(nil, string.format("%s %d/%d (%d%%)", majorMessage, (completed + 1), total, gameManager.dlcPercentComplete))
    end
end

function gameManager.displayDlcOverrideVersion(dlcVersion, isOverride)
    if dlcVersion and gameManager.currentSceneInstance then              
        local labelText
        labelText = isOverride
                        and string.format("DLC Override: %s", dlcVersion)  
                        or string.format("DLC: %s", dlcVersion)  
        echoInfo(labelText)
        local dlcLabel = cc.Label:createWithTTF(labelText, FONT_ARIAL or FONT_GILLSANS_BOLD, 10, cc.size(0,0), cc.TEXT_ALIGNMENT_CENTER, cc.VERTICAL_TEXT_ALIGNMENT_CENTER)
        dlcLabel:setAnchorPoint(cc.p(0.0,  1.0))                -- position it at the top left corner of the screen
        dlcLabel:setPosition(cc.p(10,  display.height - 10))
        gameManager.currentSceneInstance:addChild(dlcLabel)
    end
end

--  ************************************************
--  *        DLC Thread section ends               *
--  ************************************************


--  ************************************************
--  * login update message handling section starts *
--  ************************************************
function gameManager.SetLoginProgress(progress, data)
    if progress < PROGRESS_DLC then
        gameManager.currentDLCFiles = 0
    end

    ---- PREDICTS PROGRESS FIRST: ----------------
    -- local dlcComplete = 0
    -- local dlcTotal = 3  --  a guess 
    -- if progress >= PROGRESS_DLC then
    --     dlcComplete, dlcTotal = dlcManager:getDlcCounts()
    -- end
    -- local totalSteps = PROGRESS_GAME + dlcTotal
    -- local currentStep = progress + dlcComplete
    -----------------------------------------------

    local percentage = 0
    if progress == PROGRESS_START then
        percentage = 0
    elseif progress == PROGRESS_INIT then
        percentage = 10
    elseif progress == PROGRESS_SETTINGS then
        percentage = 20
    elseif data then
        percentage = 20 
        if data.total ~= 0 and data.completed ~= 0 then
            percentage = percentage + ((data.completed/data.total) * 80)
        end
    elseif progress == PROGRESS_LOGGED_IN then
        percentage = 100
    elseif progress == PROGRESS_USERDATA then
        percentage = 100
    elseif progress == PROGRESS_GAME then
        percentage = 100
    end

    gameManager.loginPercentComplete = percentage
    echoInfo("Setting login Percent Complete = %s Progress = %s", gameManager.loginPercentComplete, progress)
end

local EVENT_UPDATE_MESSAGE = "eventUpdateMessage"
function gameManager.SetUpdateMessage(majorMessage, minorMessage)
    if (majorMessage ~= nil) then
        gameManager.currentUpdateMajorMessage = majorMessage;
    end
    if (minorMessage ~= nil) then
        gameManager.currentUpdateMinorMessage = minorMessage
    end
    local event = cc.EventCustom:new(EVENT_UPDATE_MESSAGE)
    event._usedata = {major = majorMessage, minor = minorMessage}
    cc.Director:getInstance():getEventDispatcher():dispatchEvent(event)
end

function gameManager.GetUpdateMajorMessage()
    return (gameManager.currentUpdateMajorMessage or "")
end

function gameManager.GetUpdateMinorMessage()
    return (gameManager.currentUpdateMinorMessage or "")
end

function gameManager.GetLoginPercentComplete()
    return gameManager.loginPercentComplete or 0
end

function gameManager.GetDlcPercentComplete()
    return gameManager.dlcPercentComplete or 0
end

gameManager.updateListener = nil
function gameManager.SubscribeToUpdateMessages(callback)
    -- echoError("SubscribeToUpdateMessages")
    if gameManager.updateListener ~= nil then
        gameManager.UnsubscribeToUpdateMessages()
    end
    gameManager.updateListener = cc.EventListenerCustom:create(EVENT_UPDATE_MESSAGE, callback)
    eventDispatcher:addEventListenerWithFixedPriority(gameManager.updateListener, 1)    
    return gameManager.updateListener   
end

function gameManager.UnsubscribeToUpdateMessages()
    -- echoError("UnsubscribeToUpdateMessages")
    eventDispatcher:removeEventListener(gameManager.updateListener)
end



--  ************************************************
--  *  login update message handling section ends  *
--  ************************************************

function gameManager.getGameSpecificSettings()
    return gameManager.gameSpecificSettings
end


function gameManager.updateStickers()
    -- iOS only
    if device.client_platform == "ios" then
        -- iOS v10+
        if device.systemVersionGreaterThan("10", true) then 
            echoInfo("Saving Stickers")
            local fileUtils = cc.FileUtils:getInstance()

            local documentsFolder = string.gsub(device.writablePath, "[\\\\/]+$", "") .. "/"
            local stickerFolderPath = documentsFolder .. "stickers/"
            device.ensureFolderExists(stickerFolderPath)
            local stickerArray = device.getFilesInPath(stickerFolderPath, false)

            local ok,sharedPath = luaoc.callStaticMethod("AppController", "getSharedPath", nil)
            -- manually copy the files over
            dump(sharedPath, "found Shared path")
            if sharedPath ~= nil then
                sharedPath = string.gsub(sharedPath, "file://", "")
                sharedPath = sharedPath .. "Library/stickers/"
    
                -- temp hack for hard coded stickers in the /res/stickers folder for testing
                -- stickerFolderPath = cc.FileUtils:getInstance():fullPathForFilename("res/stickers/placeholder.txt")
                -- stickerFolderPath = string.gsub(stickerFolderPath, "placeholder.txt", "")
                -- device.ensureFolderExists(stickerFolderPath)
                -- dump(stickerFolderPath, "TEMP HACK STICKER PATH")
                -- stickerArray = device.getFilesInPath(stickerFolderPath, false)
    
                -- local documentsFolderReg = string.gsub(documentsFolder, "/", "//")  --  replace for regex
                -- echoInfo("Replacing %s", documentsFolderReg)
                -- echoInfo("With      %s", sharedPath)
                device.ensureFolderExists(sharedPath)
                for _, stickerFile in pairs(stickerArray) do
                    local stickerGamePath = stickerFolderPath .. stickerFile
                    local stickerSharedPath = sharedPath..stickerFile
                    if not io.exists(stickerSharedPath) then
                        echoInfo("----\nCopying sticker file from \n<--%s\n-->%s", stickerGamePath, stickerSharedPath)
                        device.copyFile(stickerGamePath, stickerSharedPath)
                    end
                end
            end
        end
    end
end

return gameManager

