local mailboxManager = {}
local scheduler = require(cc.PACKAGE_NAME .. ".scheduler")
local eventDispatcher = cc.Director:getInstance():getEventDispatcher()

mailboxManager.mailboxInfo = {}

function mailboxManager.init(serverManager, updateFreq)
	mailboxManager.mailboxInfo = mailboxManager.buildMailbox()
	mailboxManager.outwardsMailQueue = mailboxManager.buildOutboundQueue()
	mailboxManager.serverManager = serverManager;
	mailboxManager.updateFreq = updateFreq
    eventDispatcher:addEventListenerWithFixedPriority(cc.EventListenerCustom:create(MAIL_INCOMING, mailboxManager.handleMail), 1)
    mailboxManager.updateEvent = nil
    mailboxManager.paused = false
end

function mailboxManager.handleMail(event)
	mailboxManager.mailboxInfo.pushMail(event._usedata)
end

function mailboxManager.buildMailbox()
	local self = {pendingQueue = {}, processingQueue = {}, donePile = {}}
	local autoFlushHandle

	local pending = function()
		return next(self.pendingQueue) ~= nil
	end

	local function alreadyRecievedMailWithId(id)
		for k = 1,#self.pendingQueue do
            local v = self.pendingQueue[k]
			if v.id == id then return true end
		end
		for k = 1,#self.processingQueue do
            local v = self.processingQueue[k]
			if v.id == id then return true end
		end
		for k = 1,#self.donePile do
            local v = self.donePile[k]
			if v.id == id then return true end
		end
		return false
	end

	local function processMail()
		if not pending() then
			echoInfo('Nothing to process')
			return
		end
		echoInfo("Processing new mail")
		for k = 1,#self.pendingQueue do
            local mail = self.pendingQueue[k]
			table.insert(self.processingQueue, mail)
		end
		self.pendingQueue = {}
		local mailBag = {}
		for k = 1,#self.processingQueue do
            local mail = self.processingQueue[k]
			table.insert(mailBag, mail)
			table.insert(self.donePile, mail)
		end
		self.processingQueue = {}
		local event = cc.EventCustom:new(MAIL_PROCESS)
		event._usedata = mailBag
		eventDispatcher:dispatchEvent(event)
	end

	local flushMail = function()
		processMail()
	end

	local pushMail = function(mail)
		if not alreadyRecievedMailWithId(mail.id) then
			table.insert(self.pendingQueue, mail)
			if autoFlushHandle ~= nil then
				scheduler.unscheduleGlobal(autoFlushHandle)
			end
			autoFlushHandle = scheduler.performWithDelayGlobal(flushMail, 2)
		end
	end

	local printDescription = function()
		echoInfo('Inbound Queue')
		dump(self.pendingQueue, 'pendingQueue')
		dump(self.processingQueue, 'processingQueue')
		dump(self.donePile, 'donePile')
	end

	local export = {
		pending = pending,
		flushMail = flushMail,
		pushMail = pushMail,
		printDescription = printDescription
	}
	return export
end


function mailboxManager.buildOutboundQueue()
	local self = {pendingQueue = {}}

	local pending = function()
		return next(self.pendingQueue) ~= nil
	end

	local function processMail()
		if not pending() then
			echoInfo('mailboxManager: Nothing to process')
			return
		end

		local function onComplete(result)
			echoInfo("mailboxManager.processMail(outbound)=> onComplete")
		end

		local function sendItems()
			mailboxManager.serverManager.performRequest( "MailboxCreate", self.pendingQueue, nil, onComplete, true) -- endpoing, data, extraHheaders, callback, enableRetry
			self.pendingQueue = {}
		end

		sendItems()

	end

	local flushMail = function()
		processMail()
	end

	local pushMail = function(mail)
		dump(mail, "pushMail item")
		table.insert(self.pendingQueue, mail)
	end


	local printDescription = function()
		echoInfo('Outbound Queue')
		dump(self.pendingQueue, 'pendingQueue')
	end

	local export = {
		pending = pending,
		flushMail = flushMail,
		pushMail = pushMail,
		printDescription = printDescription, 
	}
	return export
end

--	tell the server that we have received and handled a mail item
function mailboxManager.acceptMail(mailId, callback)
	callback = callback or function() end
	local mailItem = {}
    mailItem["mail_item_id"] = mailId
    -- mailboxManager.serverManager.performRequest( "MailboxRead", mailItem, nil, callback, true) -- endpoint, data, extraHheaders, callback, enableRetry

    -- nerfed until we're ready:
    persistence.addPlatformTask("mailItemRead", mailItem)
    callback({code = runaway.PluginResult.kSucceed})
end

function mailboxManager.sendMail(toId, data)
	local mailItem = {}
    mailItem["recipient_id"] = toId
    mailItem["data"] = data
	mailboxManager.outwardsMailQueue.pushMail(mailItem)
end

--periodically check the server for new mail
function mailboxManager.handleMailFromServer(mailItems)
    if (mailItems) then
    	for k = 1,#mailItems do
            local mailItem = mailItems[k]
            mailboxManager.mailboxInfo.pushMail(mailItem)
        end
    end
end

function mailboxManager.resume(isRecurringUpdate)
	echoInfo("mailboxManager.resume()")
   	if (mailboxManager.updateEvent) then
		scheduler.unscheduleGlobal(mailboxManager.updateEvent)
	end
	if not LOCAL_MODE then
		--	if this is a recurring update, then go for the specified update frequency
		--	otherwise, if it is just starting up then trigger it in 10 seconds
		local frequency = isRecurringUpdate and mailboxManager.updateFreq or 10
		mailboxManager.updateEvent = scheduler.performWithDelayGlobal(mailboxManager.checkServer, frequency )
	end
end

function mailboxManager.stop()
	echoInfo("mailboxManager.stop()")
   	if (mailboxManager.updateEvent) then
		scheduler.unscheduleGlobal(mailboxManager.updateEvent)
		mailboxManager.updateEvent = nil
	end
end

function mailboxManager.checkServer()
    local function onMailReceived(result, data)
        -- dump(result, "onMailReceived.result")
        -- dump(data, "onMailReceived.data")
        if (result["code"] == runaway.PluginResult.kSucceed) then
        	if (data) then
        		mailboxManager.handleMailFromServer(data.mailItems)
        	end
        else
        --    --	if we don't get mail, then it's not the end of the world
        --    --, so we'll just try a few times and then silently give up 
        end
		mailboxManager.resume(true)	--	reschedule 
    end
    echoInfo("mailboxManager.checkServer")
    serverManager.performRequest("Mailbox", {}, nil, onMailReceived, true)
end

return mailboxManager