local persistence = {}

local scheduler = require(cc.PACKAGE_NAME .. ".scheduler")
local StateContainerFactory = require(cc.PACKAGE_NAME .. ".api.StateContainer")
local gameState = StateContainerFactory.createContainer()

--TODO
-- API wrapper and fake test calls

persistence.syncPerValidation = -1
persistence.freq = 20
persistence.count = 0
persistence.handle = nil
persistence.saveInProgress = false
persistence.paused = false
persistence.dataFn = nil
persistence.dataPrev = nil
persistence.serverManager = nil
persistence.diffPatcher = nil
persistence.successfulSaves = 0
persistence.saveSequence = 0
persistence.firstSave = true
persistence.globalOnSavedFunc = nil

function persistence.init(serverManager, dataFn, freq, syncPerValidation, globalOnSavedFunc)
	echoInfo("Initialising persistence")
	persistence.serverManager = serverManager
	persistence.dataFn = dataFn
	persistence.freq = freq
	persistence.syncPerValidation = syncPerValidation
	persistence.globalOnSavedFunc = globalOnSavedFunc


	local function gameStateListener(event)
		return event.values or {}
	end

    persistence.diffPatcher = runaway.LuaTableDiff:getInstance()
	if LOCAL_MODE then
		echoInfo("Running in fake server mode.  No diffPatching will occur")
	else
	    echoInfo("wiring up DiffPatcher singleton")
	    if (USE_DIFF_PATCHER) then
		    persistence.diffPatcher:registerEventHandler(persistence.saveDiff)
		end
    end
    
    persistence.setFirstSave()
	gameState.init(gameStateListener, "syncData", "key")
end

function persistence.cleanup()
	persistence.saveInProgress = false
end

function persistence.resume()
	persistence.paused = false
	--echoInfo("resuming persistence")

	-- ensure we don't have double ups
	if (persistence.handle) then
		scheduler.unscheduleGlobal(persistence.handle)
	end
	persistence.handle = scheduler.performWithDelayGlobal(persistence.scheduledSave, persistence.freq )
end

function persistence.scheduledSave()
	if persistence.handle then	--	only continue saving if the scheduler hasn't been deleted
		if persistence.saveInProgress then	--	if a save is currently in progress then don't bother trying to save again
			--	remember, this is just a scheduled save, so if someone has forced a save then we're pretty up to date already
			-- persistence.saveQueued = true
		else
			persistence.save()	
		end
	end
end


function persistence.save(block, preSaveToDisk)
	local data = persistence.dataFn()

	-- dump(persistence.dataPrev ,  "persistence.dataPrev")
	--dump(data ,  "persistence.dataFn")
	echoInfo("persistence.save()")

	persistence.saveInProgress = true
	persistence.count = persistence.count + 1
	print(persistence.count, persistence.syncPerValidation)
	if persistence.syncPerValidation > 0 and persistence.count >= persistence.syncPerValidation then
		echoInfo("request validation")
		persistence.validateData()
		persistence.count = 0
	end

	--Calc sync
	persistence.dataBeingSaved = clone(data)

	if preSaveToDisk then
		--	preSaveToDisk is used to save to disk prior to saving to server
		--	this handles the minimise scenario where we don't get a response from the HTTP request until the app is restored
		gameState.save(persistence.dataBeingSaved)
	end

	if (LOCAL_MODE == false) and USE_DIFF_PATCHER and (persistence.diffPatcher ~= nil) and not preSaveToDisk and not persistence.firstSave then
		local diffResult = false
		if block ~= nil and data[block] ~= nil and persistence.dataPrev ~= nil and persistence.dataPrev[block] ~= nil then
			--	block saving not entirely supported yet
		    diffResult = persistence.diffPatcher:diff(persistence.dataPrev[block], data[block]);
		else
			echoInfo("Starting full data diff")
		    -- dump(persistence.dataPrev, "persistence.dataPrev")
		    -- dump(persistence.dataBeingSaved, "persistence.dataBeingSaved")
		    diffResult = persistence.diffPatcher:diff(persistence.dataPrev, persistence.dataBeingSaved);
		end

		if not diffResult then
			echoInfo("diffPatcher failed, so just do a full save")
			local diffResult = {code = runaway.PluginResult.kSucceed, isFull = true};
			persistence.saveDiff(diffResult, persistence.dataBeingSaved)
		end
	else
		--Save locally
		if LOCAL_MODE then
			echoInfo("Skipping diff - just saving locally")
			persistence.onSave(true)
		else
			echoInfo("Skipping diff - saving to server")
			local diffResult = {code = runaway.PluginResult.kSucceed, isFull = true};
			persistence.saveDiff(diffResult, persistence.dataBeingSaved)
		end
	end
end

function persistence.saveDiff(diffResult, data)
--	echoInfo("Recieved full data diff results")

	if (type(diffResult) == "string") then
		diffResult = json.decode(diffResult)
	end
	local isDiff = not (diffResult.isFull or false);
	echoInfo("Saving to server (%s)", isDiff and "Diff" or "Full")
	-- dump(diffResult, "diffResult")
	-- dump(data, "data")

	if diffResult["code"] == runaway.PluginResult.kSucceed then

		local jsondiffs = {}
		-- if the data sent to this method is already a string, then it is the output of the diffpatcher
		--	so it needs to be decoded back to a table
		if (type(data) == "string") then
			jsondiffs = json.decode(data)
		else
			jsondiffs = data
		end

		--Save remotely

		if persistence.serverManager ~= nil then

			local platformTasksToSend = persistence._getPlatformTasks()
			--	don't bother saving empty data changes
			local hasDataToSave = ((type(jsondiffs) == "table") and not (table.isEmpty(jsondiffs)))

			if hasDataToSave or platformTasksToSend then
				if not hasDataToSave then	--	handle the scenario where there is platformTasks, but no data
					jsondiffs = {}
				end

				jsondiffs.platformTasks = platformTasksToSend

				local attempts = 0
				local retry = nil;
				local function onSaveToServer(result, data)
					echoInfo("Server.UserPost got a response")
					if result.code == HTTPREQUEST_SUCCESS then
						echoInfo("persistence.onSaveToServer = success")
						local inSequence = true
						local saveSucceeded = true
						if type(data.errors) == "table" and (not table.isEmpty(data.errors)) then
							for _, errorNum in pairs(data.errors) do
								if errorNum == 29 then	--	save out of sequence
									inSequence = false
								elseif errorNum == 32 then -- saved with no data
									local errorDetail = {}
									errorDetail[1] = "Attempt to save empty data to server"
									errorDetail[2] = json.encode(jsondiffs)
							        sendLogToServer(errorDetail)
									saveSucceeded = false
								end
							end
						end

						if inSequence and saveSucceeded then
							--	 if the server version has changed then reset after we've saved
							if data.serverVersion and gameManager.serverVersion then
								if (data.serverVersion ~= gameManager.serverVersion) then
									persistence.stop()
								    dialogManager.showDialog( 
								                SKELETON_MESSAGES.ServerUpdated.Title, 
								                SKELETON_MESSAGES.ServerUpdated.Desc,
								                true,
								                { ui.SYSTEM_DIALOG_BUTTON_OK }, 
								                gameManager.reset,
								                {hsUrl = SKELETON_MESSAGES.ServerUpdated.hsUrl});
								end
							end

							persistence.onSave(true)
						else
							echoInfo("Save is out of sequence, so performing full save")
							local fullResult = {code = runaway.PluginResult.kSucceed, isFull = true};
							persistence._platformTasksSentFailed()	--	put the platform tasks back into the queue ready for re-save
							persistence.saveDiff(fullResult, persistence.dataBeingSaved)
						end
					elseif result.code == HTTPREQUEST_ERROR_UNAUTHORIZED then
						echoInfo("persistence.onSaveToServer = Unauthorised", attempts)

						persistence._platformTasksSentFailed()	--	put the platform tasks back into the queue ready for re-save
						gameManager.handleLostSession()
					else
						echoInfo("persistence.onSaveToServer = failed %d times", attempts)
						if (attempts <= 3) then
							-- retry silently
							retry()		
						elseif (attempts <= 6) then
							--	show dialog, then retry
							persistence.handleError(retry)
						else
							--	just reset the game
							persistence._platformTasksSentFailed()	--	put the platform tasks back into the queue ready for re-save
							persistence.saveInProgress = false
							persistence.handleError(nil)
						end
						--persistence.onSave(false)
					end
				end

				local params = {}
				if isDiff then
					persistence.saveSequence = persistence.saveSequence + 1
				else
					persistence.saveSequence = 0
				end
				params["save-seq"] = persistence.saveSequence
				params["is-diff"] = isDiff and "y" or "n"
				retry = function()
					attempts = attempts + 1
					echoInfo("PerformRequest(UserPost)-> Sequence #%d", persistence.saveSequence)
					persistence.serverManager.performRequest("UserPost", jsondiffs, params, onSaveToServer)
				end

				retry();
			else
				echoInfo("Skipping save to server - no data")
				persistence.onSave(true)
			end
		else
			echoInfo("Skipping save to server - local mode")
			persistence.onSave(true)
		end

	else
		echoInfo("DiffPatcher Failed")
		persistence.onSave(false)
	end
end

function persistence.onSave(isSuccess)
--	echoInfo("Save to server complete")
	
	if isSuccess then	
		persistence.firstSave = false
		echoInfo("Saving to disk")
		--	only update the saved state if it was successful
		--		(the exception to this is when we force the data to save (eg: on minimise)
		--		this is because the app goes to sleep before getting here.
		persistence._platformTasksSent()
		persistence.dataBeingSaved.platformTasks = nil 	-- so these aren't stored on the filesystem	
		gameState.save(persistence.dataBeingSaved)
		--dump(persistence.dataBeingSaved, "persistence Data saved to disk")
		persistence.dataPrev = persistence.dataBeingSaved	--	this was previously cloned, so no need to do it here
		persistence.dataBeingSaved = nil
		persistence.successfulSaves = persistence.successfulSaves + 1
		echoInfo("successful Saves = %s", persistence.successfulSaves)

		-- if we've been given a callback method in the persistence.force method then call it, and then forget the callback.
		if (type(persistence.onSaveCompleteCallback) == "function") then
			persistence.onSaveCompleteCallback()
			persistence.onSaveCompleteCallback = nil
		end
		if (type (persistence.globalOnSavedFunc) == "function") then
			persistence.globalOnSavedFunc()
		end
	else
		persistence.successfulSaves = 0;
		echoInfo("persistence.save failed")
	end
	persistence.saveInProgress = false

	if persistence.saveQueued then
		persistence.saveQueued = false
		echoInfo("Persistence save was queued, re-running")
		if persistence.saveQueuedParams then
	 		local resumeAfterForce = persistence.saveQueuedParams.resumeAfterForce
	 		local block = persistence.saveQueuedParams.block
	 		local preSaveToDisk = persistence.saveQueuedParams.preSaveToDisk
	 		local onSaveCompleteCallback = persistence.saveQueuedParams.onSaveCompleteCallback
	 		persistence.saveQueuedParams = nil
	 		persistence.force(resumeAfterForce, block, preSaveToDisk, onSaveCompleteCallback)
	 	else
			persistence.save()
		end
	else
		if not persistence.paused then
			echoInfo("Resuming persistence after save")
			persistence.resume()	--	re-schedule the next save
		end
	end
end

--	forces the local data to be saved.  only use this if we know that our copy of the data is identical to the the server data (such as upon a restore session by deviceid).
function persistence.setDataSyncedWithServer()
	--this saves the data locally, and saves our record of the state of the data on the server (to avoid large diffPatches)
	echoInfo("marking data as synced with server")
	local data = persistence.dataFn(true)	--	true == ignore beforeSaveFunc - this just gets the raw data
	gameState.save(data)
	persistence.dataPrev = clone(data)
end

function persistence.setFirstSave()
	persistence.saveSequence = 0
	persistence.dataPrev = nil
	persistence.firstSave = true
end

function persistence.stop()
	persistence.paused = true -- 15th October code change to fix Lua Error.
	scheduler.unscheduleGlobal(persistence.handle)
	persistence.handle = nil
end

function persistence.load()
	echoInfo("persistence.load()")
	return gameState.load()
end

function persistence.force(resumeAfterForce, block, preSaveToDisk, onSaveCompleteCallback)
	--	preSaveToDisk is used to save to disk prior to saving to server
	--	this handles the minimise scenario where we don't get a response from the HTTP request until the app is restored
	--	it also forces a full (non-diff) save
	if persistence.saveInProgress then
		persistence.saveQueued = true
		persistence.saveQueuedParams = {resumeAfterForce = resumeAfterForce, block = block, preSaveToDisk = preSaveToDisk, onSaveCompleteCallback = onSaveCompleteCallback}
	else
		if (onSaveCompleteCallback ~= nil) then
			persistence.onSaveCompleteCallback = onSaveCompleteCallback;
		end
		echoInfo("persistence.force()")
		-- persistence.stop()	--	stop the scheduler
		-- if not resumeAfterForce then
		-- 	persistence.paused = true
		-- end
		persistence.save(block, preSaveToDisk)
	end
end


function persistence.isSyncInProgress()
	return persistence.saveInProgress
end

function persistence.validateData()
	echoInfo("persistence.validateData()")
end

function persistence.handleError(onAcceptErrorCallback)
	local wasPaused = persistence.paused
	persistence.stop()
	local function onAcceptError()
		if type(onAcceptErrorCallback) == "function" then
			--	if an onAcceptedCallback was provided then we need to resume persistence if it was previously enabled
			persistence.paused = wasPaused
			onAcceptErrorCallback()
		else
			--	otherwise just reset the game
			gameManager.reset()
		end
	end

    dialogManager.showDialog(
                SKELETON_MESSAGES.NetworkFailedPersistence.Title,
                SKELETON_MESSAGES.NetworkFailedPersistence.Desc,
                true,
                { ui.SYSTEM_DIALOG_BUTTON_OK }, 
                onAcceptError,
                {hsUrl = SKELETON_MESSAGES.NetworkFailedPersistence.hsUrl});
end

function persistence.isOffline()
end

function persistence.backup()
end

function persistence.delete()
	gameState.delete()
	persistence.stop()
	persistence.setFirstSave()
end

function persistence.isLocalData()
	return cc.FileUtils:getInstance():isFileExist(gameState.getStateContainerPath())
end

function persistence.addPlatformTask(taskType, taskData)
	persistence.platformTasks = persistence.platformTasks or {}
	persistence.platformTasks[taskType] = persistence.platformTasks[taskType] or {} 
	table.insert(persistence.platformTasks[taskType], taskData)
	dump(persistence.platformTasks, "persistence.addPlatformTask->platformTasks", 10)
end

function persistence._getPlatformTasks()
	--	grab the current list of tasks from the queue to send them
	persistence.platformTasksSent = clone(persistence.platformTasks or {})
	persistence.platformTasks = {}
	if not table.isEmpty(persistence.platformTasksSent) then
		dump(persistence.platformTasksSent, "platformTasks sending")
		return persistence.platformTasksSent
	else
		echoInfo("no platformTasks to send")
		return nil
	end
end

function persistence._platformTasksSent()
	--	sending succeeded, so we'll clear the sent list
	persistence.platformTasksSent = nil
end

function persistence._platformTasksSentFailed()
	--	sending tasks to the server failed, so we'll just put them back in the queue
	if persistence.platformTasksSent then
		dump(persistence.platformTasks, "PlatformTasks before re-merge", 10)

		for taskBucketKey, taskBucket in pairs(persistence.platformTasksSent) do
			local taskList = persistence.platformTasks[taskBucketKey]
			if taskList then
				-- merge the lists
				for _, taskItem in pairs(taskBucket) do
					table.insert(taskList, taskItem)
				end
			else
				persistence.platformTasks[taskBucketKey] = taskBucket
			end
		end
		dump(persistence.platformTasks, "PlatformTasks after re-merge", 10)
		persistence.platformTasksSent = nil
	end
end


return persistence