local Server = class("Server")
local scheduler = require(cc.PACKAGE_NAME .. ".scheduler")
local socketUrl = require("socket.socket.url")

local m_valid_http_methods = {
	GET = true,
	HEAD = true,
	POST = true,
	PUT = true,
	DELETE = true
}

-- m_endpoints


local ALLCHARS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
local SALT_LENGTH = 16
function Server:createRandomSalt()
    local newSalt = ""
    for i = 1, tonumber(SALT_LENGTH) do
      local r = math.random(#ALLCHARS)
      newSalt = newSalt.. string.sub(ALLCHARS, r, r)
    end
    return newSalt
end

function Server:ctor()
	echoInfo("Server:ctor()")
	self.m_endpoints = {}
end


-- @param endpoints is a table containing the URLs where the Service Provider exposes its endpoints
--   each endpoint is either a string (its url, the method is POST by default) or a table, with the url in the array part and the method 
--   in the 'method' field.
function Server:setEndPoints(endpoints)

	for k,v in pairs(endpoints or {}) do
		if type(v) == "table" then
			self.m_endpoints[k] = { url = v[1], method = v.method }
		else
			self.m_endpoints[k] = { url = v, method = HTTPREQUEST_POST }
		end
	end
end

--
-- Sets / gets m_endpoints
function Server:SetEndpoints(value)
	self.m_endpoints = value
end
function Server:GetEndpoints()
	return self.m_endpoints
end

function Server:GetEndpoint(key)
	return self.m_endpoints[key]
end

--
-- Encryption
function Server:SetEncryptionCipher(cipher)
	self.m_cipher = cipher
end

function Server:SetEncryptionKey(key)
	self.m_key = key
end

function Server:SetEncryptionIV(iv)
	self.m_iv = iv
end

function Server:encrypt(data)
	local salt = self:createRandomSalt()
	local content = crypto.encrypt(self.m_cipher, crypto.pad(data), self.m_key, salt)
	local result = ":" .. salt .. ":" .. crypto.encodeBase64(content)
	return result
end

function Server:decrypt(raw)
	local salt = self.m_iv 	--	default
	if (string.sub(raw, 1, 1) == ":") and (string.sub(raw, 18, 18) == ":") then
		salt = string.sub(raw, 2, 17)
		raw = string.sub(raw, 19)
	end
	local decoded = crypto.decodeBase64(raw)
	local result
	if (decoded == nil) and string.len(raw) ~= 0 then
		--	not actually encoded!!?!
		result = raw
	else
		result = crypto.unpad(crypto.decrypt(self.m_cipher, decoded, self.m_key, salt))
	end
	return result
end


function Server:PerformRequest(method, url, arguments, headers, callback)
	echoInfo("Server:PerformRequest")

	scheduler.performWithDelayGlobal(function()
		callback(nil, {timestamp=os.time(), gameVersion=1})
	end, 1)
end

function Server:checkIsJson(headers)
	if type(headers) == "table" then
		for k, header in pairs(headers) do
			if string.find(header, "application/json") then
				return true
			end
		end
	end

	return false
end

function Server:getValues(headers, response_body)
	local values = {}
	if response_body ~= nil then
		if(self:checkIsJson(headers)) then
			values = json.decode(response_body)
		else
			for key, value in string.gmatch(response_body, "([^&=]+)=([^&=]*)&?" ) do
				--print( ("key=%s, value=%s"):format(key, value) )
				values[key] = socketUrl.unescape(value)
			end
		end
	end

	return values
end

--
-- Like URL-encoding, but following OAuth's specific semantics
function Server:oauth_encode(val)
	return val:gsub('[^-._~a-zA-Z0-9]', function(letter)
		return string.format("%%%02x", letter:byte()):upper()
	end)
end

--
-- COnverts params into / spearted string.
function Server:convertParamsToUrl(arguments)
    local body = {}
    for k,v in pairs(arguments) do
        body[#body + 1] = "/" .. tostring(v)
    end
    return table.concat(body, "/")
end

--
-- Sets / gets oauth_token
function Server:SetToken( value)
	self.m_oauth_token = value
end
function Server:GetToken()
	return self.m_oauth_token
end

--
-- Sets / gets oauth_token_secret
function Server:SetTokenSecret(value)
	self.m_oauth_token_secret = value
end
function Server:GetTokenSecret()
	return self.m_oauth_token_secret
end

--
-- Sets / gets oauth_verifier
function Server:SetVerifier(value)
	self.m_oauth_verifier = value
end
function Server:GetVerifier()
	return self.m_oauth_verifier
end

--
-- Sets / gets oauth_access_token
function Server:SetAccessToken(value)
	self.m_oauth_access_token = value
end
function Server:GetAccessToken()
	return self.m_oauth_access_token
end

--
-- Sets / gets oauth_token
function Server:SetRefreshToken(value)
	self.m_oauth_refresh_token = value
end
function Server:GetRefeshToken()
	return self.m_oauth_refresh_token
end


return Server