
local M = {}

function M.addOrdinalIndicator(num)
   return num..M.getOrdinalIndicator(num)
end
------------------------------

function M.getOrdinalIndicator(num)
   local mod = tonumber(num)%10
   local hundredMod = tonumber(num)%100
   local indicators = {'st', 'nd', 'rd', 'th', 'th', 'th', 'th', 'th', 'th', 'th', 'th', 'th', 'th'}
   local result = ""
   if indicators[mod] and not indicators[hundredMod] then
      return indicators[mod] 
   else
      if indicators[hundredMod] then
         return indicators[hundredMod]
      else
         return 'th'
      end
   end
end
------------------------------

function M.getSignOfPointToLine(pt, lineA, lineB)
   local determinant = (lineB.x - lineA.x) * (pt.y - lineA.y) - (lineB.y - lineA.y) * (pt.x - lineA.x)
   local sign = determinant / math.abs(determinant)
   return sign
end
------------------------------

function M.getSaveDir(absolute)
   if absolute then 
      return string.format("%sother",runaway.FileHelpers:getTmpPath())..'/'
   else
      if device.client_platform == 'android' or device.client_platform == 'amazon' then
         return 'other/' else
         return '../tmp/other/'
      end
   end
end
------------------------------

function M.resizeSprite(sprite, targetSize)
   local currentSize = sprite:getContentSize()

   local targetScale = cc.size(0,0)
   targetScale.width = targetSize.width/currentSize.width
   targetScale.height = targetSize.height/currentSize.height

   local scale = (targetScale.width > targetScale.height) and targetScale.height or targetScale.width
   sprite:setScale(scale)
   return scale
end
------------------------------

function M.convert3Bto4F(color3b)
   local color4f = cc.c4f(color3b.r/255, color3b.g/255, color3b.b/255, 1.0)
   return color4f
end
------------------------------

function M.getAllDescendants(parent)
   local descendants = {}
   local function gatherDescendants(node)
      local children = node:getChildren()
      for _,child in pairs(children) do
         table.insert(descendants, child)
         if #child:getChildren() > 0 then
            gatherDescendants(child)
         end
      end
   end
   gatherDescendants(parent)
   return descendants
end
------------------------------

function M.packCirclesIntoCircle(circles, args)
   
   -- Input = an indexed table of radii
   -- Output = and indexed table of offsets

   local args = args or {}

   local points = {}
   local radius = args.radius or 100
   local distBetweenPoints = args.distBetweenPoints or 10
   local numConcentricCircles = math.ceil(radius/distBetweenPoints)
   local startPoint = args.startPoint or cc.p(0,0)
   local vThreshold = args.vThreshold or false
   
   points[1] = {dist = 0, position = startPoint, radius = 0}
   local ix = 2

   for i = 1, numConcentricCircles do
      local length = (distBetweenPoints * i)
      local pointsPerCircle = (length * 2 * math.pi)/distBetweenPoints
      local pointStep = 360/pointsPerCircle
      for ii = 0, pointsPerCircle do
         local point = M.getPointInDirection(startPoint, length, ii * pointStep)
         points[ix] = {dist = 0, position = point, radius = 0}
         ix = ix+1
      end
   end

   if vThreshold then
      local tempPoints = {}
      for i = 1, #points do
         if points[i].position.y > vThreshold then
            tempPoints[#tempPoints + 1] = points[i]
         end
      end
      points = tempPoints
   end

   return M.packCircles(circles, points)
end
------------------------------

function M.packCirclesIntoTriangle(leaderRad, circles)

   -- Packs a bunch of circles of different sizes into a
   -- triangle with the first ones being at the left point of the triangle.

   -- Input = an indexed table of radii
   -- Output = and indexed table of offsets

   -- Creating the Triangle of points: The higher the
   -- number of columns and the lower the distance between them,
   -- the more closely the circles will be packed together.

   -- Make sure that currentColumn * triangleDist == radius of the leader
   -- fish, or else position 1 in the queue will clip the leader

   local startPoint = cc.p(0,0)
   local leaderRad = leaderRad or 0
   leaderRad = leaderRad + 10

   local triangle = {}
   local triangleColumns = 30
   local triangleDist = 10
   local currentColumn = math.ceil(leaderRad/triangleDist)
   local ix = 1

   for row = currentColumn, triangleColumns, 1 do
      for index = 1, row, 1 do 
         local testPoint = cc.p(currentColumn * triangleDist, (index - (row/2)) * triangleDist)
         triangle[ix] = {dist = 0, position = testPoint, radius = 0}
         ix = ix + 1
         if index == row then
            currentColumn = currentColumn + 1
         end
      end
   end
   for index = 1,#triangle do
      local trianglePoint = triangle[index]
      local dist = M.squaredDist(startPoint, trianglePoint.position)
      trianglePoint.dist = dist
   end
   table.sort(triangle, function(a,b) return a.dist < b.dist end)

   return M.packCircles(circles, triangle)
end
------------------------------

function M.packCircles(circles, points)

   local tempCircles = {}

   -- For each of the circles, iterate through the table of points and find a point
   -- where the circle doesn't intersect with any other circles already placed
   -- in the triangle:

   local circleIndex = 1
   for circleIndex = 1, #circles do
      local circleRadius = circles[circleIndex]
      local validPointForCircle = nil

      local targetPointIndex = 1
      for targetPointIndex = 1, #points do
         local targetPoint = points[targetPointIndex]
         local stillGood = true

         if targetPoint.radius == 0 then

            local testPointIndex = 1
            for testPointIndex = 1, #points do
               local testPoint = points[testPointIndex]

               if testPointIndex ~= targetPointIndex then
                  -- Distance squared calculation
                  local diffX = targetPoint.position.x - testPoint.position.x
                  local diffY = targetPoint.position.y - testPoint.position.y
                  local distSq = (diffX * diffX) + (diffY * diffY)

                  local circleTotalRadiusSq = (circleRadius + testPoint.radius)
                  circleTotalRadiusSq = circleTotalRadiusSq * circleTotalRadiusSq

                  if testPoint.radius > 0 and
                     (distSq < circleTotalRadiusSq) then
                     stillGood = false
                     break
                  end
               end
            end
         end
         if targetPoint.radius > 0 then stillGood = false end
         if stillGood then
            validPointForCircle = targetPoint
            break
         end
      end

      if validPointForCircle then
         tempCircles[circleIndex] = {radius = circleRadius, position = validPointForCircle.position}

         local targetPointIndex = 1
         for targetPointIndex = 1, #points do
            local targetPoint = points[targetPointIndex]
            if targetPoint.position.x == validPointForCircle.position.x and
               targetPoint.position.y == validPointForCircle.position.y then
               targetPoint.radius = circleRadius
            end
         end
      else
         tempCircles[circleIndex] = {radius = circleRadius, position = cc.p(0,0)}
      end
   end

   return tempCircles
end
------------------------------

function M.convertShadowOffsetFromPointToSize(point)

   -- Enable shadow on cc.label:createWithTTF only accepts cc.size as
   -- an offset which doesn't include the content scale factor. Because
   -- of this, we're manually converting an offset as a point into 
   -- a size which does include the content scale factor:

   local offsetAsSize = cc.size((1/display.contentScaleFactor) * point.x, (1/display.contentScaleFactor) * point.y)
   return offsetAsSize
end
------------------------------

function M.purgeTempDirectory()

   -- Call this function if you need to empty the tmp folder
   -- this folder contains the stored pngs, map.json and
   -- any other files that you want to store between sessions
   -- but clear between different versions of the game.

   echoInfo('Purging tmp/other directory...')

   require "lfs"

   local filePath = M.getSaveDir(true)
   local exists = cc.FileUtils:getInstance():isFileExist(filePath)

   if exists then
      for file in lfs.dir(filePath) do
          if (file ~= ".") and (file ~= "..") then
              device.removePath(filePath.."/"..file)
          end
      end
   end
end
------------------------------

function M.splitString(str, delim, maxNb)
   -- Eliminate bad cases...
   if string.find(str, delim) == nil then
      return { str }
   end
   if maxNb == nil or maxNb < 1 then
      maxNb = 0   -- No limit
   end
   local result = {}
   local pat = "(.-)" .. delim .. "()"
   local nb = 0
   local lastPos
   for part, pos in string.gfind(str, pat) do
      nb = nb + 1
      result[nb] = part
      lastPos = pos
      if nb == maxNb then break end
   end
   -- Handle the last field
   if nb ~= maxNb then
      result[nb + 1] = string.sub(str, lastPos)
   end
   return result
end
------------------------------
-- Adapted from: http://alienryderflex.com/polygon/. Returns if point 'p' is
-- inside polygon 'polygon'. can be concave or convex, passed in as a table of cc.points.

function M.isPointInPolygon(p, polygon)
   local minX = 0
   local maxX = 0
   local minY = 0
   local maxY = 0
   for i = 1, #polygon, 1 do
      local q = polygon[i]
      minX = math.min(q.x, minX)
      maxX = math.max(q.x, maxX)
      minY = math.min(q.y, minY)
      maxY = math.max(q.y, maxY)
   end

   if p.x < minX or p.x > maxX or p.y < minY or p.y > maxY then
      return false
   end

   local intersections = 0
   local j = #polygon
   for i=1, #polygon, 1 do
      local test1 = polygon[i].y > p.y and 1 or 0
      local test2 = polygon[j].y > p.y and 1 or 0
      if test1 ~= test2 then  
         if p.x < ((polygon[j].x - polygon[i].x) * (p.y - polygon[i].y) / (polygon[j].y - polygon[i].y ) + polygon[i].x) then
            intersections = intersections + 1
         end
      end
      j = i
   end
   
   if intersections%2 == 0 then
      return false
   end
   return true
end
------------------------------

function M.squaredDist(point1, point2)
   local diffX = point1.x - point2.x
   local diffY = point1.y - point2.y
   local dist = (diffX * diffX) + (diffY * diffY)
   return dist
end
------------------------------

function M.angleToVector(angle)
   local angleInRads = angle * DEGREES_TO_RADIANS
   local vec = cc.p(math.cos(angleInRads), -math.sin(angleInRads))
   return vec
end
------------------------------

function M.isPointInRect(point, rect)
   if point.x > rect.x and point.x < rect.x + rect.width then
      if point.y > rect.y and point.y < rect.y + rect.height then
         return true
      end
   end
   return false
end
------------------------------

function M.doRectsIntersect(rect1, rect2)
   local bottomLeft = cc.p(rect1.x, rect1.y)
   local topLeft = cc.p(rect1.x, rect1.y + rect1.height)
   local bottomRight = cc.p(rect1.x + rect1.width, rect1.y)
   local topRight = cc.p(rect1.x + rect1.width, rect1.y + rect1.height)
   if M.isPointInRect(bottomLeft, rect2) then  return true end
   if M.isPointInRect(topLeft, rect2) then     return true end
   if M.isPointInRect(bottomRight, rect2) then return true end
   if M.isPointInRect(topRight, rect2) then    return true end

   return false
end
------------------------------

function M.segIntersection(x1, y1, x2, y2, x3, y3, x4, y4)
   -- Line segment intersection
   local bx = x2 - x1; 
   local by = y2 - y1; 
   local dx = x4 - x3; 
   local dy = y4 - y3;
   local b_dot_d_perp = bx * dy - by * dx;
   if(b_dot_d_perp == 0) then
      return nil;
   end
   local cx = x3 - x1;
   local cy = y3 - y1;
   local t = (cx * dy - cy * dx) / b_dot_d_perp;
   if(t < 0 or t > 1) then
      return nil;
   end
   local u = (cx * by - cy * bx) / b_dot_d_perp;
   if(u < 0 or u > 1) then
      return nil;
   end
   return x1+t*bx, y1+t*by
end
------------------------------

function M.isPointOnScreen(point)
   local onScreen = true
   if point.x < 0 or point.x > display.width or
      point.y < 0 or point.y > display.height then
      onScreen = false
   end
   return onScreen
end
------------------------------

function M.getPointInDirection(startPoint, length, direction)
   local spiralX = length * math.cos(direction * DEGREES_TO_RADIANS)
   local spiralY = length * math.sin(direction * DEGREES_TO_RADIANS)
   return cc.p(startPoint.x + spiralX, startPoint.y + spiralY)
end
------------------------------

function M.getCenterOfPolygon(polygon)
   local rect = cc.rect(99999,99999,0,0)
   for i = 1,#polygon do
      local point = polygon[i]
      if point.x < rect.x then
         rect.x = point.x
      end
      if point.x > rect.width then
         rect.width = point.x
      end
      if point.y < rect.y then
         rect.y = point.y
      end
      if point.y > rect.height then
         rect.height = point.y
      end
   end
   local centerPoint = cc.p(rect.width - rect.x, rect.height - rect.y)
   centerPoint.x = centerPoint.x / 2
   centerPoint.y = centerPoint.y / 2
   centerPoint.x = rect.x + centerPoint.x
   centerPoint.y = rect.y + centerPoint.y
   return centerPoint, rect
end
------------------------------

function M.getRandomPointAroundOrigin(origin, variance)
   return cc.p(origin.x + (-(variance/2) + (variance * math.random())), origin.y + (-(variance/2 + (variance * math.random()))))
end
------------------------------

function M.addCommas(amount)
  local formatted = tostring(amount)
  local k
  while true do  
    formatted, k = string.gsub(formatted, "^(-?%d+)(%d%d%d)", '%1,%2')
    if (k==0) then
      break
    end
  end
  return formatted
end
------------------------------

function M.removeCommas(numString)
   local noCommasString = string.gsub(numString, ",", "")
   return tonumber(noCommasString)
end
------------------------------

function M.createTextureFromSprite(sprite)
   local size = sprite:getContentSize()
   local renderTexture = cc.RenderTexture:create(size.width, size.height)

   sprite:setPosition(size.width/2, size.height/2)
   renderTexture:begin()
      renderTexture.visit(sprite)
   renderTexture:endToLua()

   return renderTexture:getSprite():getTexture()
end
------------------------------

function M.timeUnitsToShortString(units)
   s = ''
   for i = 1,#units do
      local unit = units[i]
      if s == '' then
         if i == #units then
            s = string.format('0:%02d', unit)
         elseif unit ~= 0 then
            s = string.format('%d', unit)
         end
      else
         s = string.format('%s:%02d', s, unit)
      end
   end
   return s
end
------------------------------------

function M.capitaliseFirstLettersInString(str)
   local capitalisedStr = str:gsub("(%l)(%w*)", function(a,b) return string.upper(a)..b end)
   return capitalisedStr
end

function M.secondsToTimeLeftString(t)
   local days     = math.floor(t / SECONDS_IN_DAY)
   local daysrem  = math.floor(t % SECONDS_IN_DAY)
   local hrs      = math.floor(daysrem / SECONDS_IN_HOUR)
   local hrsrem   = math.floor(daysrem % SECONDS_IN_HOUR)
   local mins     = math.floor(hrsrem / SECONDS_IN_MINUTE)
   local minsrem  = math.floor(hrsrem % SECONDS_IN_MINUTE)
   local sec      = math.floor(minsrem)

   return M.timeUnitsToShortString({days,hrs,mins,sec})
end
------------------------------------

function M.secondsToTimeLeftShortString(t)
   local daysmod  = t % SECONDS_IN_DAY
   local hrsmod   = t % SECONDS_IN_HOUR
   local minsmod  = t % SECONDS_IN_MINUTE

   if     ( daysmod == 0 ) then
      t = t / SECONDS_IN_DAY
      return t .. 'd' 
   elseif ( hrsmod == 0 ) then
      t = t / SECONDS_IN_HOUR
      return t .. 'h'
   elseif ( minsmod == 0 ) then
      t = t / SECONDS_IN_MINUTE
      return t .. 'm'
   else
      return t .. 's'
   end

   return  ""
end
------------------------------------

function M.secondsLeftToDaysHrs(t)
   local days     = math.floor(t / SECONDS_IN_DAY)
   local daysrem  = math.floor(math.abs(t) % SECONDS_IN_DAY)
   local hrs      = math.floor(daysrem / SECONDS_IN_HOUR)

   local s = ''
   if t > 0 and days > 0 then
      local suffix = days == 1 and '' or 's'
      s = s..days..' day'..suffix
   end

   if t > 0 and hrs > 0 then
      if days > 0 then s = s..' ' end
      local suffix = hrs == 1 and '' or 's'
      s = s..hrs..' hr'..suffix
   end

   if s == '' then s = '0 hrs' end
   return s
end
------------------------------------

-- also returns if it is using days
function M.secondsLeftToDaysOrHrs(t)
   local days     = math.floor(t / SECONDS_IN_DAY)
   local daysrem  = math.floor(math.abs(t) % SECONDS_IN_DAY)
   local hrs      = math.floor(daysrem / SECONDS_IN_HOUR)

   if t > 0 and days > 0 then
      local suffix = days == 1 and '' or 's'
      return days..' day'..suffix, true
   end

   if t > 0 and hrs > 0 then
      local suffix = hrs == 1 and '' or 's'
      return hrs..' hr'..suffix, false
   end

   return '0 hrs', false
end
------------------------------------

-- also returns if it is using days
function M.secondsToDaysOrHours(t)
   local days     = math.floor(t / SECONDS_IN_DAY)
   local daysrem  = math.floor(math.abs(t) % SECONDS_IN_DAY)
   local hrs      = math.floor(daysrem / SECONDS_IN_HOUR)

   if t > 0 and days > 0 then
      local suffix = days == 1 and '' or 's'
      return days..' day'..suffix, true
   end

   if t > 0 and hrs > 0 then
      local suffix = hrs == 1 and '' or 's'
      return hrs..' hour'..suffix, false
   end

   return '0 hours', false
end
------------------------------------

function M.secondsToTimeLeftMediumString(t)
   local days     = math.floor(t / SECONDS_IN_DAY)
   local daysrem  = math.floor(t % SECONDS_IN_DAY)
   local hrs      = math.floor(daysrem / SECONDS_IN_HOUR)
   local hrsrem   = math.floor(daysrem % SECONDS_IN_HOUR)
   local mins     = math.floor(hrsrem / SECONDS_IN_MINUTE)
   local minsrem  = math.floor(hrsrem % SECONDS_IN_MINUTE)
   local sec      = math.floor(minsrem)

   local s = "";
   if (days > 0) then
      s = s .. days .. "d"
   end

   if ( hrs > 0 ) then
      if (days > 0) then s = s .. " " end
      s = s .. hrs .. "h"
   end

   if ( mins > 0 ) then
      if ( days > 0 or hrs > 0) then s = s .. " " end
      s = s .. mins .. "m"
   end

   if ( sec > 0 ) then
      if ( days > 0 or hrs > 0 or mins > 0 ) then s = s .. " " end
      s = s .. sec .. "s"
   end
      
   return s
end
------------------------------------

function M.secondsToTimeLeftLongString(t)
   local days     = math.floor(t / SECONDS_IN_DAY)
   local daysrem  = math.floor(t % SECONDS_IN_DAY)
   local hrs      = math.floor(daysrem / SECONDS_IN_HOUR)
   local hrsrem   = math.floor(daysrem % SECONDS_IN_HOUR)
   local mins     = math.floor(hrsrem / SECONDS_IN_MINUTE)
   local minsrem  = math.floor(hrsrem % SECONDS_IN_MINUTE)
   local sec      = math.floor(minsrem)

   local s = "";
   if (days > 0) then
      s = s .. days .. "day" .. ( (days ~= 1) and "s" or "" )
   end

   if ( hrs > 0 ) then
      if (days > 0) then s = s .. " " end
      s = s .. hrs .. "hr" .. ( (hrs ~= 1) and "s" or "" )
   end

   if ( mins > 0 ) then
      if ( days > 0 or hrs > 0) then s = s .. " " end
      s = s .. mins .. "min" .. ( (mins ~= 1) and "s" or "" )
   end

   if ( sec > 0 ) then
      if ( days > 0 or hrs > 0 or mins > 0 ) then s = s .. " " end
      s = s .. sec .. "sec" .. ( (sec ~= 1) and "s" or "" )
   end
      
   return s
end
------------------------------------

function M.getScaleFromNodeToGrandparent(node, grandparent, offsetPos)
   if node == nil then return 1, 1 end
   if grandparent == nil then return node:getScaleX(), node:getScaleY() end

   local offsetPos = offsetPos or cc.p(0, 0)
   local parent = node:getParent()
   local nodeScreenPos = node:getParent():convertToWorldSpace(cc.p(node:getPosition()))
   local relativePos = cc.p(nodeScreenPos.x - offsetPos.x, nodeScreenPos.y - offsetPos.y)
   local scaleX = node:getScaleX()
   local scaleY = node:getScaleY()
   while parent ~= grandparent do
      if parent == nil then return end

      scaleX = scaleX * parent:getScaleX()
      scaleY = scaleY * parent:getScaleY()
      parent = parent:getParent()
   end
   return scaleX, scaleY, relativePos
end
------------------------------

function M.getQuerystringParams(url)
   params = {}
   if type(url) == "string" then
      local splitUrl = utils.splitString(url, '?')
      if splitUrl[2] ~= nil then
         local allParams = utils.splitString(splitUrl[2], "&")
         for _, param in pairs(allParams) do
            local splitParam = utils.splitString(param, "=")
            if #splitParam == 2 then
               params[splitParam[1]] = splitParam[2]
            end
         end
      end
   end
   return params
end

return M
